from __future__ import annotations

from typing import Any, TypeAlias, cast

import pytest

pytest.importorskip("litestar")

from litestar.connection import ASGIConnection, Request, WebSocket
from litestar.datastructures.state import State
from litestar.enums import ScopeType
from litestar.types.asgi_types import HTTPScope, WebSocketScope

import diwire._internal.integrations.litestar as litestar_integration
from diwire import Container
from diwire.exceptions import DIWireIntegrationError

LitestarConnection: TypeAlias = ASGIConnection[object, object, object, State]
LitestarRequest: TypeAlias = Request[object, object, State]
LitestarWebSocket: TypeAlias = WebSocket[object, object, State]


def _build_http_scope(path: str = "/") -> HTTPScope:
    return cast(
        "HTTPScope",
        {
            "type": ScopeType.HTTP,
            "asgi": {"version": "3.0", "spec_version": "2.3"},
            "http_version": "1.1",
            "method": "GET",
            "scheme": "http",
            "path": path,
            "raw_path": path.encode(),
            "query_string": b"",
            "headers": [],
            "client": ("testclient", 50000),
            "server": ("testserver", 80),
            "state": {},
        },
    )


def _build_websocket_scope(path: str = "/ws") -> WebSocketScope:
    return cast(
        "WebSocketScope",
        {
            "type": ScopeType.WEBSOCKET,
            "asgi": {"version": "3.0", "spec_version": "2.3"},
            "scheme": "ws",
            "path": path,
            "raw_path": path.encode(),
            "query_string": b"",
            "headers": [],
            "client": ("testclient", 50000),
            "server": ("testserver", 80),
            "subprotocols": [],
            "state": {},
        },
    )


def _request_for_scope(path: str = "/") -> Request[object, object, State]:
    return Request(_build_http_scope(path))


def _websocket_for_scope(path: str = "/ws") -> WebSocket[object, object, State]:
    async def _receive() -> Any:
        return {"type": "websocket.disconnect", "code": 1000}

    async def _send(_message: Any) -> None:
        return None

    return WebSocket(
        _build_websocket_scope(path),
        receive=_receive,
        send=_send,
    )


def test_get_connection_raises_when_context_is_missing() -> None:
    with pytest.raises(DIWireIntegrationError, match="Connection context not available"):
        litestar_integration.get_connection(object, object, object, State)


def test_get_request_raises_when_current_connection_is_not_http_request() -> None:
    token = litestar_integration._connection_context.set(_websocket_for_scope("/websocket"))
    try:
        with pytest.raises(DIWireIntegrationError, match="not HTTP Request"):
            litestar_integration.get_request(object, object, State)
    finally:
        litestar_integration._connection_context.reset(token)


def test_get_websocket_raises_when_current_connection_is_not_websocket() -> None:
    token = litestar_integration._connection_context.set(_request_for_scope("/request"))
    try:
        with pytest.raises(DIWireIntegrationError, match="not WebSocket"):
            litestar_integration.get_websocket(object, object, State)
    finally:
        litestar_integration._connection_context.reset(token)


def test_get_request_returns_current_http_request_connection() -> None:
    request = _request_for_scope("/request")
    token = litestar_integration._connection_context.set(request)
    try:
        resolved = litestar_integration.get_request(object, object, State)
        assert resolved is request
    finally:
        litestar_integration._connection_context.reset(token)


def test_get_websocket_returns_current_websocket_connection() -> None:
    websocket = _websocket_for_scope("/websocket")
    token = litestar_integration._connection_context.set(websocket)
    try:
        resolved = litestar_integration.get_websocket(object, object, State)
        assert resolved is websocket
    finally:
        litestar_integration._connection_context.reset(token)


@pytest.mark.asyncio
async def test_request_context_middleware_sets_http_request_context() -> None:
    observed_path: str | None = None

    async def app(_scope: Any, _receive: Any, _send: Any) -> None:
        nonlocal observed_path
        observed_path = litestar_integration.get_request(object, object, State).url.path

    middleware = litestar_integration.RequestContextMiddleware()
    wrapped_app = middleware(app)
    scope = _build_http_scope("/middleware/http")

    async def receive() -> Any:
        return {"type": "http.request", "body": b"", "more_body": False}

    async def send(_message: Any) -> None:
        return None

    await wrapped_app(scope, receive, send)

    assert observed_path == "/middleware/http"


@pytest.mark.asyncio
async def test_request_context_middleware_sets_websocket_context() -> None:
    observed_path: str | None = None

    async def app(_scope: Any, _receive: Any, _send: Any) -> None:
        nonlocal observed_path
        observed_path = litestar_integration.get_websocket(object, object, State).url.path

    middleware = litestar_integration.RequestContextMiddleware()
    wrapped_app = middleware(app)
    scope = _build_websocket_scope("/middleware/websocket")

    async def receive() -> Any:
        return {"type": "websocket.connect"}

    async def send(_message: Any) -> None:
        return None

    await wrapped_app(scope, receive, send)

    assert observed_path == "/middleware/websocket"


@pytest.mark.asyncio
async def test_request_context_middleware_resets_context_after_exception() -> None:
    async def app(_scope: Any, _receive: Any, _send: Any) -> None:
        assert (
            litestar_integration.get_request(object, object, State).url.path == "/middleware/error"
        )
        msg = "handler failed"
        raise RuntimeError(msg)

    middleware = litestar_integration.RequestContextMiddleware()
    wrapped_app = middleware(app)
    scope = _build_http_scope("/middleware/error")

    async def receive() -> Any:
        return {"type": "http.request", "body": b"", "more_body": False}

    async def send(_message: Any) -> None:
        return None

    with pytest.raises(RuntimeError, match="handler failed"):
        await wrapped_app(scope, receive, send)

    with pytest.raises(DIWireIntegrationError, match="Connection context not available"):
        litestar_integration.get_connection(object, object, object, State)


def test_add_request_context_registers_factories_for_container_resolution() -> None:
    container = Container()
    litestar_integration.add_request_context(container)

    request = _request_for_scope("/container/request")
    request_token = litestar_integration._connection_context.set(request)
    try:
        resolved_connection = container.resolve(LitestarConnection)
        resolved_request = container.resolve(LitestarRequest)
        assert resolved_connection is request
        assert resolved_request is request
    finally:
        litestar_integration._connection_context.reset(request_token)

    websocket = _websocket_for_scope("/container/websocket")
    websocket_token = litestar_integration._connection_context.set(websocket)
    try:
        resolved_websocket = container.resolve(LitestarWebSocket)
        assert resolved_websocket is websocket
    finally:
        litestar_integration._connection_context.reset(websocket_token)



from __future__ import annotations

import inspect
from contextvars import ContextVar
from typing import Any, Generic, TypeVar, cast

import pytest

from diwire import (
    Container,
    DependencyRegistrationPolicy,
    Injected,
    Lifetime,
    Maybe,
    MissingPolicy,
    Provider,
    Scope,
    resolver_context,
)
from diwire.exceptions import DIWireDependencyNotRegisteredError


class _MaybeDependency:
    pass


class _MissingNestedDependency:
    pass


class _BrokenService:
    def __init__(self, dependency: _MissingNestedDependency) -> None:
        self.dependency = dependency


_DEFAULT_SENTINEL = object()
_INJECT_DEFAULT_SENTINEL = object()
_SHIFT_DEFAULT_SENTINEL = object()
_POSITIONAL_ONLY_DEFAULT_SENTINEL = object()


class _OptionalDefaultConsumer:
    def __init__(self, dependency: Maybe[_MaybeDependency] = _DEFAULT_SENTINEL) -> None:
        self.dependency = dependency


class _OptionalNoneConsumer:
    def __init__(self, dependency: Maybe[_MaybeDependency]) -> None:
        self.dependency = dependency


class _RequestScopedDependency:
    pass


class _ShiftedOptionalResult:
    def __init__(self, dependency: object, value: int) -> None:
        self.dependency = dependency
        self.value = value


def _build_shifted_optional_result(
    dependency: Maybe[_MaybeDependency] = _SHIFT_DEFAULT_SENTINEL,
    value: int = 5,
) -> _ShiftedOptionalResult:
    return _ShiftedOptionalResult(dependency=dependency, value=value)


class _PositionalOnlyOptionalResult:
    def __init__(self, dependency: object, value: int) -> None:
        self.dependency = dependency
        self.value = value


def _build_positional_only_optional_result(
    dependency: Maybe[_MaybeDependency] = _POSITIONAL_ONLY_DEFAULT_SENTINEL,
    value: int = 5,
    /,
) -> _PositionalOnlyOptionalResult:
    return _PositionalOnlyOptionalResult(dependency=dependency, value=value)


def _build_varargs_optional_result(*values: int) -> tuple[int, ...]:
    return values


def _build_kwargs_optional_result(**values: int) -> dict[str, int]:
    return values


T = TypeVar("T")


class _MaybeBox(Generic[T]):
    pass


class _MaybeBoxImpl(_MaybeBox[T]):
    def __init__(self, type_arg: type[T]) -> None:
        self.type_arg = type_arg


def _strict_container() -> Container:
    return Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )


def test_resolve_maybe_returns_none_for_unregistered_dependency_in_strict_mode() -> None:
    container = _strict_container()

    assert container.resolve(Maybe[_MaybeDependency]) is None


def test_resolve_maybe_returns_registered_dependency_value() -> None:
    container = _strict_container()
    dependency = _MaybeDependency()
    container.add_instance(dependency, provides=_MaybeDependency)

    assert container.resolve(Maybe[_MaybeDependency]) is dependency


def test_resolve_maybe_propagates_nested_registration_errors() -> None:
    container = _strict_container()
    container.add(_BrokenService, provides=_BrokenService)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(Maybe[_BrokenService])


def test_missing_maybe_dependency_uses_constructor_default() -> None:
    container = _strict_container()
    container.add(_OptionalDefaultConsumer, provides=_OptionalDefaultConsumer)

    resolved = container.resolve(_OptionalDefaultConsumer)

    assert resolved.dependency is _DEFAULT_SENTINEL


def test_missing_maybe_dependency_without_default_injects_none() -> None:
    container = _strict_container()
    container.add(_OptionalNoneConsumer, provides=_OptionalNoneConsumer)

    resolved = container.resolve(_OptionalNoneConsumer)

    assert resolved.dependency is None


def test_registered_maybe_dependency_overrides_constructor_default() -> None:
    container = _strict_container()
    dependency = _MaybeDependency()
    container.add_instance(dependency, provides=_MaybeDependency)
    container.add(_OptionalDefaultConsumer, provides=_OptionalDefaultConsumer)

    resolved = container.resolve(_OptionalDefaultConsumer)

    assert resolved.dependency is dependency


def test_inject_maybe_dependency_uses_default_when_missing_and_overrides_when_registered() -> None:
    container = _strict_container()

    @resolver_context.inject
    def handler(
        dependency: Injected[Maybe[_MaybeDependency]] = _INJECT_DEFAULT_SENTINEL,
    ) -> object:
        return dependency

    injected_handler = cast("Any", handler)
    assert injected_handler() is _INJECT_DEFAULT_SENTINEL

    dependency = _MaybeDependency()
    container.add_instance(dependency, provides=_MaybeDependency)
    assert injected_handler() is dependency


def test_inject_scope_inference_uses_inner_dependency_scope_for_maybe() -> None:
    container = _strict_container()
    container.add(
        _RequestScopedDependency,
        provides=_RequestScopedDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @resolver_context.inject(auto_open_scope=True)
    def handler(
        dependency: Injected[Maybe[_RequestScopedDependency]],
    ) -> _RequestScopedDependency | None:
        return dependency

    injected_handler = cast("Any", handler)
    resolved = injected_handler()
    assert isinstance(resolved, _RequestScopedDependency)


def test_resolve_maybe_closed_generic_uses_open_generic_registration_match() -> None:
    container = _strict_container()
    container.add(_MaybeBoxImpl, provides=_MaybeBox)

    resolved = container.resolve(Maybe[_MaybeBox[int]])

    assert isinstance(resolved, _MaybeBoxImpl)
    assert resolved.type_arg is int


def test_resolve_maybe_with_contextvar_backed_provider_uses_current_value() -> None:
    current_value_var: ContextVar[int] = ContextVar("maybe_current_value", default=3)

    def read_current_value() -> int:
        return current_value_var.get()

    container = _strict_container()
    container.add_factory(
        read_current_value,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        assert request_scope.resolve(Maybe[int]) == 3
        token = current_value_var.set(7)
        try:
            assert request_scope.resolve(Maybe[int]) == 7
        finally:
            current_value_var.reset(token)


def test_resolve_maybe_provider_token_keeps_provider_semantics() -> None:
    container = _strict_container()

    dependency_provider = container.resolve(Maybe[Provider[_MaybeDependency]])
    assert callable(dependency_provider)
    with pytest.raises(DIWireDependencyNotRegisteredError):
        dependency_provider()


def test_optional_union_has_no_special_resolution_semantics() -> None:
    container = _strict_container()

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(_MaybeDependency | None)


def test_missing_optional_positional_or_keyword_dependency_uses_keywords_for_following_arguments() -> (
    None
):
    signature = inspect.signature(_build_shifted_optional_result)
    container = _strict_container()
    container.add_instance(99, provides=int)
    container.add_factory(
        _build_shifted_optional_result,
        provides=_ShiftedOptionalResult,
        dependencies={
            Maybe[_MaybeDependency]: signature.parameters["dependency"],
            int: signature.parameters["value"],
        },
    )

    resolved = container.resolve(_ShiftedOptionalResult)

    assert resolved.dependency is _SHIFT_DEFAULT_SENTINEL
    assert resolved.value == 99


def test_missing_optional_positional_only_dependency_omits_subsequent_positional_only_arguments() -> (
    None
):
    signature = inspect.signature(_build_positional_only_optional_result)
    container = _strict_container()
    container.add_instance(99, provides=int)
    container.add_factory(
        _build_positional_only_optional_result,
        provides=_PositionalOnlyOptionalResult,
        dependencies={
            Maybe[_MaybeDependency]: signature.parameters["dependency"],
            int: signature.parameters["value"],
        },
    )

    resolved = container.resolve(_PositionalOnlyOptionalResult)

    assert resolved.dependency is _POSITIONAL_ONLY_DEFAULT_SENTINEL
    assert resolved.value == 5


def test_missing_optional_varargs_dependency_resolves_to_empty_tuple_literal() -> None:
    signature = inspect.signature(_build_varargs_optional_result)
    container = _strict_container()
    container.add_factory(
        _build_varargs_optional_result,
        provides=tuple[int, ...],
        dependencies={
            Maybe[_MaybeDependency]: signature.parameters["values"],
        },
    )

    resolved = container.resolve(tuple[int, ...])

    assert resolved == ()


def test_missing_optional_kwargs_dependency_resolves_to_empty_dict_literal() -> None:
    signature = inspect.signature(_build_kwargs_optional_result)
    container = _strict_container()
    container.add_factory(
        _build_kwargs_optional_result,
        provides=dict[str, int],
        dependencies={
            Maybe[_MaybeDependency]: signature.parameters["values"],
        },
    )

    resolved = container.resolve(dict[str, int])

    assert resolved == {}



from __future__ import annotations

from dataclasses import dataclass
from types import TracebackType

from aiohttp import web
from typing_extensions import Self

_cleanup_paths: list[str] = []


def latest_cleanup_path() -> str | None:
    if not _cleanup_paths:
        return None
    return _cleanup_paths[-1]


@dataclass
class RequestBasedService:
    request: web.Request

    def work(self) -> str:
        return f"Request path: {self.request.path}"


@dataclass
class WebSocketBasedService:
    request: web.Request

    def work(self) -> str:
        return f"WebSocket path: {self.request.path}"


@dataclass
class CMService:
    request: web.Request

    def work(self) -> str:
        return "CMService working"

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        _cleanup_paths.append(self.request.path)



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, cast

import pytest

from diwire import Container, LockMode


class InstanceService:
    pass


class ConcreteService:
    pass


class FactoryService:
    pass


class GeneratorService:
    pass


class ContextManagerService:
    pass


class OverrideConcreteService:
    pass


class OverrideFactoryService:
    pass


class OverrideGeneratorService:
    pass


class OverrideContextManagerService:
    pass


class OverrideDecoratedFactoryService:
    pass


def test_container_default_lock_mode_auto() -> None:
    container = Container()
    container.add(ConcreteService)

    assert container._providers_registrations.get_by_type(ConcreteService).lock_mode == "auto"


def test_registration_default_from_container_applies_to_all_non_instance_registrations() -> None:
    def build_factory() -> FactoryService:
        return FactoryService()

    def build_generator() -> Generator[GeneratorService, None, None]:
        try:
            yield GeneratorService()
        finally:
            pass

    @contextmanager
    def build_context_manager() -> Generator[ContextManagerService, None, None]:
        yield ContextManagerService()

    container = Container(lock_mode=LockMode.NONE)
    container.add_instance(InstanceService())
    container.add(ConcreteService)
    container.add_factory(build_factory)
    container.add_generator(build_generator)
    container.add_context_manager(build_context_manager)

    assert (
        container._providers_registrations.get_by_type(InstanceService).lock_mode is LockMode.NONE
    )
    assert (
        container._providers_registrations.get_by_type(ConcreteService).lock_mode is LockMode.NONE
    )
    assert container._providers_registrations.get_by_type(FactoryService).lock_mode is LockMode.NONE
    assert (
        container._providers_registrations.get_by_type(GeneratorService).lock_mode is LockMode.NONE
    )
    assert (
        container._providers_registrations.get_by_type(ContextManagerService).lock_mode
        is LockMode.NONE
    )


def test_registration_override_lock_mode_takes_precedence() -> None:
    def build_factory() -> OverrideFactoryService:
        return OverrideFactoryService()

    def build_generator() -> Generator[OverrideGeneratorService, None, None]:
        try:
            yield OverrideGeneratorService()
        finally:
            pass

    @contextmanager
    def build_context_manager() -> Generator[OverrideContextManagerService, None, None]:
        yield OverrideContextManagerService()

    def build_decorated_factory() -> OverrideDecoratedFactoryService:
        return OverrideDecoratedFactoryService()

    container = Container(lock_mode=LockMode.NONE)
    container.add_factory(build_factory, lock_mode=LockMode.THREAD)
    container.add(OverrideConcreteService, lock_mode=LockMode.ASYNC)
    container.add_generator(build_generator, lock_mode=LockMode.NONE)
    container.add_context_manager(
        build_context_manager,
        lock_mode=LockMode.ASYNC,
    )

    container.add_factory(
        build_decorated_factory,
        provides=OverrideDecoratedFactoryService,
        lock_mode=LockMode.THREAD,
    )

    assert (
        container._providers_registrations.get_by_type(OverrideConcreteService).lock_mode
        is LockMode.ASYNC
    )
    assert (
        container._providers_registrations.get_by_type(OverrideFactoryService).lock_mode
        is LockMode.THREAD
    )
    assert (
        container._providers_registrations.get_by_type(OverrideGeneratorService).lock_mode
        is LockMode.NONE
    )
    assert (
        container._providers_registrations.get_by_type(OverrideDecoratedFactoryService).lock_mode
        is LockMode.THREAD
    )
    assert (
        container._providers_registrations.get_by_type(OverrideContextManagerService).lock_mode
        is LockMode.ASYNC
    )


def test_register_instance_rejects_lock_mode_argument() -> None:
    container = Container()

    with pytest.raises(TypeError, match="lock_mode"):
        cast("Any", container.add_instance)(
            instance=InstanceService(),
            lock_mode=LockMode.THREAD,
        )



from __future__ import annotations

import ast
import hashlib
import importlib.machinery
import importlib.util
import linecache
import re
import sys
import traceback
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType

import pytest


@dataclass(frozen=True, slots=True)
class CodeBlock:
    doc_path: Path
    header_line: int
    code: str


_FENCE_BEGIN_RE = re.compile(r"^```python\s*$")
_FENCE_END_RE = re.compile(r"^```\s*$")


def _find_repo_root(start: Path) -> Path:
    for candidate in (start, *start.parents):
        if (candidate / "pyproject.toml").is_file() and (candidate / "src" / "diwire").is_dir():
            return candidate
    msg = f"Could not locate repository root from {start}"
    raise AssertionError(msg)


def _extract_python_code_blocks(doc_path: Path) -> list[CodeBlock]:
    text = doc_path.read_text(encoding="utf-8")
    lines = text.splitlines()

    blocks: list[CodeBlock] = []
    i = 0
    while i < len(lines):
        if _FENCE_BEGIN_RE.match(lines[i]) is None:
            i += 1
            continue

        header_line = i + 1
        i += 1
        code_lines: list[str] = []
        while i < len(lines) and _FENCE_END_RE.match(lines[i]) is None:
            code_lines.append(lines[i])
            i += 1

        code = "\n".join(code_lines).rstrip()
        blocks.append(CodeBlock(doc_path=doc_path, header_line=header_line, code=code))

        while i < len(lines) and _FENCE_END_RE.match(lines[i]) is None:
            i += 1
        if i < len(lines):
            i += 1

    return blocks


def _build_script(blocks: list[CodeBlock], repo_root: Path) -> tuple[str, list[tuple[int, str]]]:
    script_lines: list[str] = []
    markers: list[tuple[int, str]] = []
    rel_path = blocks[0].doc_path.relative_to(repo_root) if blocks else "<unknown>"

    for index, block in enumerate(blocks, start=1):
        marker = f"# --- {rel_path}:block:{index} ---"
        markers.append((len(script_lines) + 1, marker))
        script_lines.append(marker)
        if block.code:
            script_lines.extend(block.code.splitlines())
        script_lines.append("")

    return "\n".join(script_lines), markers


def _closest_marker(markers: list[tuple[int, str]], failing_line: int) -> str | None:
    best: str | None = None
    for marker_line, marker_text in markers:
        if marker_line <= failing_line:
            best = marker_text
        else:
            break
    return best


def _execution_failure_message(
    *, doc_path: Path, repo_root: Path, markers: list[tuple[int, str]], exc: BaseException
) -> str:
    extracted = traceback.extract_tb(exc.__traceback__)
    failing_line: int | None = None
    for frame in reversed(extracted):
        if frame.filename == str(doc_path):
            failing_line = frame.lineno
            break

    marker = _closest_marker(markers, failing_line) if failing_line is not None else None
    marker_note = f"\n{marker}" if marker else ""
    rel_path = doc_path.relative_to(repo_root)
    return f"README snippet execution failed: {rel_path}\n{type(exc).__name__}: {exc}{marker_note}"


def _module_name_for_doc(doc_path: Path) -> str:
    digest = hashlib.sha256(str(doc_path).encode("utf-8")).hexdigest()[:12]
    return f"diwire_repo_readme_code_blocks_{digest}"


def _module_for_exec(*, module_name: str, doc_path: Path) -> ModuleType:
    module = ModuleType(module_name)
    module.__file__ = str(doc_path)

    loader = importlib.machinery.SourceFileLoader(module_name, str(doc_path))
    spec = importlib.util.spec_from_loader(module_name, loader)
    if spec is None:
        msg = f"Could not create module spec for {doc_path}"
        raise AssertionError(msg)

    module.__loader__ = loader
    module.__spec__ = spec
    return module


REPO_ROOT = _find_repo_root(Path(__file__).resolve())
README_PATH = REPO_ROOT / "README.md"

_OPTIONAL_README_DEPS: frozenset[str] = frozenset({"fastapi", "starlette"})


def test_repo_readme_python_code_blocks_execute() -> None:
    blocks = _extract_python_code_blocks(README_PATH)
    if not blocks:
        return

    script, markers = _build_script(blocks, repo_root=REPO_ROOT)
    ast.parse(script, filename=str(README_PATH))

    module_name = _module_name_for_doc(README_PATH)
    module = _module_for_exec(module_name=module_name, doc_path=README_PATH)
    readme_path_text = str(README_PATH)
    script_lines = [f"{line}\n" for line in script.splitlines()]
    previous_linecache_entry = linecache.cache.get(readme_path_text)
    linecache.cache[readme_path_text] = (len(script), None, script_lines, readme_path_text)

    sys.modules[module_name] = module
    try:
        compiled = compile(script, filename=str(README_PATH), mode="exec")
        exec(compiled, module.__dict__)  # noqa: S102
    except ModuleNotFoundError as exc:
        missing = (exc.name or "").split(".", 1)[0]
        if missing in _OPTIONAL_README_DEPS:
            pytest.skip(f"README.md requires optional dependency: {missing}")
        msg = _execution_failure_message(
            doc_path=README_PATH,
            repo_root=REPO_ROOT,
            markers=markers,
            exc=exc,
        )
        raise AssertionError(msg) from exc
    except Exception as exc:
        msg = _execution_failure_message(
            doc_path=README_PATH,
            repo_root=REPO_ROOT,
            markers=markers,
            exc=exc,
        )
        raise AssertionError(msg) from exc
    finally:
        sys.modules.pop(module_name, None)
        if previous_linecache_entry is None:
            linecache.cache.pop(readme_path_text, None)
        else:
            linecache.cache[readme_path_text] = previous_linecache_entry



from __future__ import annotations

import uvicorn
from litestar import Litestar, get, websocket
from litestar.connection import Request, WebSocket
from litestar.datastructures.state import State
from litestar.exceptions import WebSocketDisconnect

from diwire import Container, Injected, Scope, resolver_context
from diwire.integrations.litestar import RequestContextMiddleware, add_request_context
from tests.e2e.litestar.setup.config import LitestarE2ESettings
from tests.e2e.litestar.setup.services import (
    CMService,
    RequestBasedService,
    WebSocketBasedService,
    latest_cleanup_path,
)

# diwire resolves dependencies via runtime type hints; populate module globals so
# forward references like `Request[object, object, State]` can be evaluated.
globals()["Request"] = Request
globals()["WebSocket"] = WebSocket
globals()["State"] = State


@get("/health")
async def health() -> str:
    return "OK"


@get("/services/request-based")
@resolver_context.inject(scope=Scope.REQUEST)
async def request_based_service(service: Injected[RequestBasedService]) -> str:
    return service.work()


@get("/services/cm-service")
@resolver_context.inject(scope=Scope.REQUEST)
async def cm_service(service: Injected[CMService]) -> str:
    return service.work()


@get("/services/cm-service/cleanup")
async def cm_service_cleanup() -> str:
    cleanup_path = latest_cleanup_path()
    return f"Cleanup path: {cleanup_path}"


@websocket("/services/request-based-websocket")
@resolver_context.inject(scope=Scope.REQUEST)
async def websocket_request_based_service(
    socket: WebSocket[object, object, State],
    service: Injected[WebSocketBasedService],
) -> None:
    await socket.accept()
    try:
        _ = await socket.receive_text()
    except WebSocketDisconnect:
        return
    await socket.send_text(service.work())
    await socket.close()


app = Litestar(
    route_handlers=[
        health,
        request_based_service,
        cm_service,
        cm_service_cleanup,
        websocket_request_based_service,
    ],
    middleware=[RequestContextMiddleware()],
)


def main() -> None:
    container = Container()
    add_request_context(container)
    container.add_context_manager(CMService, scope=Scope.REQUEST)

    settings = container.resolve(LitestarE2ESettings)
    uvicorn.run(app, host=settings.host, port=settings.port)


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections.abc import Awaitable, Callable
from contextvars import ContextVar

from aiohttp import web

from diwire import Container, Lifetime
from diwire.exceptions import DIWireIntegrationError

_request_context: ContextVar[web.Request | None] = ContextVar(
    "diwire_aiohttp_request_context",
    default=None,
)


def get_request() -> web.Request:
    request = _request_context.get()
    if request is None:
        msg = (
            "Request context not available. Ensure request_context_middleware is installed via "
            "app = web.Application(middlewares=[request_context_middleware])."
        )
        raise DIWireIntegrationError(msg)
    return request


@web.middleware
async def request_context_middleware(
    request: web.Request,
    handler: Callable[[web.Request], Awaitable[web.StreamResponse]],
) -> web.StreamResponse:
    token = _request_context.set(request)
    try:
        return await handler(request)
    finally:
        _request_context.reset(token)


def add_request_context(container: Container) -> None:
    container.add_factory(
        get_request,
        provides=web.Request,
        lifetime=Lifetime.TRANSIENT,
    )



"""Focused example: ``DIWireInvalidRegistrationError``."""

from __future__ import annotations

from typing import Any, cast

from diwire import Container
from diwire.exceptions import DIWireInvalidRegistrationError


def main() -> None:
    container = Container()

    try:
        container.add_instance(object(), provides=cast("Any", None))
    except DIWireInvalidRegistrationError as error:
        error_name = type(error).__name__

    print(f"invalid_reg={error_name}")  # => invalid_reg=DIWireInvalidRegistrationError


if __name__ == "__main__":
    main()



"""Focused example: strict mode without concrete autoregistration."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container, DependencyRegistrationPolicy, MissingPolicy
from diwire.exceptions import DIWireDependencyNotRegisteredError


class Dependency:
    pass


@dataclass(slots=True)
class Root:
    dependency: Dependency


def main() -> None:
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )

    try:
        container.resolve(Root)
    except DIWireDependencyNotRegisteredError as error:
        error_name = type(error).__name__

    print(f"strict_missing={error_name}")  # => strict_missing=DIWireDependencyNotRegisteredError


if __name__ == "__main__":
    main()



"""Focused example: ``DIWireAsyncDependencyInSyncContextError``."""

from __future__ import annotations

from diwire import Container
from diwire.exceptions import DIWireAsyncDependencyInSyncContextError


class AsyncDependency:
    pass


async def provide_async_dependency() -> AsyncDependency:
    return AsyncDependency()


def main() -> None:
    container = Container()
    container.add_factory(provide_async_dependency, provides=AsyncDependency)

    try:
        container.resolve(AsyncDependency)
    except DIWireAsyncDependencyInSyncContextError as error:
        error_name = type(error).__name__

    print(f"async_in_sync={error_name}")  # => async_in_sync=DIWireAsyncDependencyInSyncContextError


if __name__ == "__main__":
    main()



from __future__ import annotations

import importlib
from collections.abc import MutableMapping
from types import SimpleNamespace
from typing import Any

import pytest

pytest.importorskip("starlette")

from starlette.requests import Request, StateT
from starlette.websockets import WebSocket

import diwire._internal.integrations.fastapi as fastapi_integration
from diwire.exceptions import DIWireIntegrationError


def _build_http_scope(path: str = "/") -> dict[str, Any]:
    return {
        "type": "http",
        "asgi": {"version": "3.0", "spec_version": "2.3"},
        "http_version": "1.1",
        "method": "GET",
        "scheme": "http",
        "path": path,
        "raw_path": path.encode(),
        "query_string": b"",
        "headers": [],
        "client": ("testclient", 50000),
        "server": ("testserver", 80),
        "state": {},
    }


def _build_websocket_scope(path: str = "/ws") -> dict[str, Any]:
    return {
        "type": "websocket",
        "asgi": {"version": "3.0", "spec_version": "2.3"},
        "scheme": "ws",
        "path": path,
        "raw_path": path.encode(),
        "query_string": b"",
        "headers": [],
        "client": ("testclient", 50000),
        "server": ("testserver", 80),
        "subprotocols": [],
        "state": {},
    }


def _request_for_scope(path: str = "/") -> Request:
    return Request(_build_http_scope(path))


def _websocket_for_scope(path: str = "/ws") -> WebSocket:
    async def _receive() -> MutableMapping[str, Any]:
        return {"type": "websocket.disconnect", "code": 1000}

    async def _send(_message: MutableMapping[str, Any]) -> None:
        return None

    return WebSocket(
        _build_websocket_scope(path),
        receive=_receive,
        send=_send,
    )


def test_get_connection_raises_when_context_is_missing() -> None:
    with pytest.raises(DIWireIntegrationError, match="Connection context not available"):
        fastapi_integration.get_connection(dict)


def test_get_request_raises_when_current_connection_is_not_http_request() -> None:
    token = fastapi_integration._connection_context.set(_websocket_for_scope("/websocket"))
    try:
        with pytest.raises(DIWireIntegrationError, match="not HTTP Request"):
            fastapi_integration.get_request(dict)
    finally:
        fastapi_integration._connection_context.reset(token)


def test_get_websocket_typed_raises_when_current_connection_is_not_websocket() -> None:
    token = fastapi_integration._connection_context.set(_request_for_scope("/request"))
    try:
        with pytest.raises(DIWireIntegrationError, match="not WebSocket"):
            fastapi_integration._get_websocket_typed(dict)
    finally:
        fastapi_integration._connection_context.reset(token)


def test_get_websocket_typed_returns_current_websocket_connection() -> None:
    websocket = _websocket_for_scope("/websocket")
    token = fastapi_integration._connection_context.set(websocket)
    try:
        resolved = fastapi_integration._get_websocket_typed(dict)
        assert resolved is websocket
    finally:
        fastapi_integration._connection_context.reset(token)


def test_get_websocket_untyped_raises_when_current_connection_is_not_websocket() -> None:
    token = fastapi_integration._connection_context.set(_request_for_scope("/request"))
    try:
        with pytest.raises(DIWireIntegrationError, match="not WebSocket"):
            fastapi_integration._get_websocket_untyped()
    finally:
        fastapi_integration._connection_context.reset(token)


def test_get_websocket_selector_chooses_typed_path_when_state_typevar_is_present(
    monkeypatch: Any,
) -> None:
    module = importlib.reload(fastapi_integration)
    monkeypatch.setattr(
        module.WebSocket,
        "__orig_bases__",
        (SimpleNamespace(__args__=(StateT,)),),
        raising=False,
    )

    reloaded = importlib.reload(module)

    assert reloaded.get_websocket is reloaded._get_websocket_typed


def test_get_websocket_selector_falls_back_to_untyped_path_on_orig_bases_errors(
    monkeypatch: Any,
) -> None:
    module = importlib.reload(fastapi_integration)
    monkeypatch.setattr(module.WebSocket, "__orig_bases__", (), raising=False)

    reloaded = importlib.reload(module)

    assert reloaded.get_websocket is reloaded._get_websocket_untyped



from __future__ import annotations

from dataclasses import dataclass
from uuid import uuid4

import pytest

pytest.importorskip("celery")

from celery import Celery, Task

from diwire import Container, Injected, Lifetime, Scope, resolver_context


@dataclass(frozen=True)
class _AdderService:
    def add(self, left: int, right: int) -> int:
        return left + right


@dataclass(frozen=True)
class _ScopedToken:
    value: str


def _build_scoped_token() -> _ScopedToken:
    return _ScopedToken(value=uuid4().hex)


@pytest.fixture()
def celery_app() -> Celery:
    container = Container()
    container.add(_AdderService, provides=_AdderService)
    container.add_factory(
        _build_scoped_token,
        provides=_ScopedToken,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    app = Celery("diwire-tests", broker="memory://", backend="cache+memory://")
    app.conf.update(
        task_always_eager=True,
        task_store_eager_result=True,
        task_eager_propagates=True,
        task_serializer="json",
        result_serializer="json",
        accept_content=["json"],
    )

    @app.task(name="diwire.tests.add")
    @resolver_context.inject(scope=Scope.REQUEST)
    def add_task(left: int, right: int, adder: Injected[_AdderService]) -> int:
        return adder.add(left, right)

    @app.task(name="diwire.tests.scope")
    @resolver_context.inject(scope=Scope.REQUEST)
    def scoped_token_task(
        first: Injected[_ScopedToken],
        second: Injected[_ScopedToken],
    ) -> dict[str, str]:
        return {"first": first.value, "second": second.value}

    @app.task(bind=True, name="diwire.tests.bound")
    @resolver_context.inject(scope=Scope.REQUEST)
    def bound_task(
        task: Task,
        value: int,
        adder: Injected[_AdderService],
    ) -> str:
        return f"{task.name}:{adder.add(value, 1)}"

    return app


def test_celery_task_resolves_injected_dependencies(celery_app: Celery) -> None:
    result = celery_app.tasks["diwire.tests.add"].delay(4, 6)
    assert result.get(timeout=5.0) == 10


def test_celery_task_scoped_dependency_reused_per_task_and_isolated_between_tasks(
    celery_app: Celery,
) -> None:
    first = celery_app.tasks["diwire.tests.scope"].delay().get(timeout=5.0)
    second = celery_app.tasks["diwire.tests.scope"].delay().get(timeout=5.0)

    assert first["first"] == first["second"]
    assert second["first"] == second["second"]
    assert first["first"] != second["first"]


def test_celery_bound_task_keeps_self_argument_and_injects_dependencies(celery_app: Celery) -> None:
    result = celery_app.tasks["diwire.tests.bound"].delay(9)
    assert result.get(timeout=5.0) == "diwire.tests.bound:10"



from __future__ import annotations

import typer

from diwire import Container, Injected, Lifetime, Scope, resolver_context
from diwire.integrations.typer import add_typer_context
from tests.e2e.typer.setup.services import ContextInfoService

app = typer.Typer()


container = Container()
add_typer_context(container)
container.add(
    ContextInfoService,
    scope=Scope.REQUEST,
    lifetime=Lifetime.SCOPED,
)


@app.command("context-direct")
@resolver_context.inject(scope=Scope.REQUEST)
def context_direct(
    ctx: typer.Context,
    resolved_ctx: Injected[typer.Context],
) -> None:
    info_name = resolved_ctx.info_name or "<missing>"
    same = "same" if resolved_ctx is ctx else "different"
    typer.echo(f"{info_name}:{same}")


@app.command("context-service")
@resolver_context.inject(scope=Scope.REQUEST)
def context_service(
    ctx: typer.Context,
    service: Injected[ContextInfoService],
) -> None:
    typer.echo(service.describe(ctx))


def main() -> None:
    app()


if __name__ == "__main__":
    main()



"""Focused example: closed generic override beats open template."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

from diwire import Container

T = TypeVar("T")


class IBox(Generic[T]):
    pass


@dataclass(slots=True)
class Box(IBox[T]):
    type_arg: type[T]


class _SpecialIntBox(IBox[int]):
    pass


def main() -> None:
    container = Container()
    container.add(Box, provides=IBox)
    container.add(_SpecialIntBox, provides=IBox[int])

    resolved = container.resolve(IBox[int])
    print(f"override={type(resolved).__name__}")  # => override=_SpecialIntBox


if __name__ == "__main__":
    main()



import uvicorn
from fastapi import FastAPI
from starlette.websockets import WebSocket, WebSocketDisconnect

from diwire import Container, Injected, Scope, resolver_context
from diwire.integrations.fastapi import RequestContextMiddleware, add_request_context
from tests.e2e.fastapi.setup.config import FastAPIE2ESettings
from tests.e2e.fastapi.setup.services import (
    CMService,
    RequestBasedService,
    RequestWithCustomStateService,
    WebSocketBasedService,
)

app = FastAPI()
app.add_middleware(RequestContextMiddleware)


@app.get("/health")
async def health() -> str:
    return "OK"


@app.get("/services/request-based")
@resolver_context.inject(scope=Scope.REQUEST)
async def request_based_service(service: Injected[RequestBasedService]) -> str:
    return service.work()


@app.get("/services/request-based-with-custom-state")
@resolver_context.inject(scope=Scope.REQUEST)
async def request_based_service_with_custom_state(
    service: Injected[RequestWithCustomStateService],
) -> str:
    return service.work()


@app.websocket("/services/request-based-with-custom-state")
@resolver_context.inject(scope=Scope.SESSION)
async def websocket_request_based_service_with_custom_state(
    ws: WebSocket,
    service: Injected[WebSocketBasedService],
) -> None:
    await ws.accept()
    while True:
        try:
            _data = await ws.receive_text()
        except WebSocketDisconnect:
            break
        await ws.send_text(service.work())


@app.get("/services/cm-service")
@resolver_context.inject(scope=Scope.REQUEST)
async def cm_service(service: Injected[CMService]) -> str:
    return service.work()


def main() -> None:
    container = Container()
    add_request_context(container)
    container.add_context_manager(CMService, scope=Scope.REQUEST)

    settings = container.resolve(FastAPIE2ESettings)
    uvicorn.run(app, host=settings.host, port=settings.port)


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import inspect
import threading
from collections.abc import AsyncGenerator, Callable, Generator
from concurrent.futures import ThreadPoolExecutor
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from contextvars import ContextVar
from dataclasses import dataclass, field
from types import TracebackType
from typing import Annotated, Any, Generic, TypeVar, cast

import pytest

from diwire import (
    Component,
    Container,
    DependencyRegistrationPolicy,
    Injected,
    Lifetime,
    LockMode,
    MissingPolicy,
    Scope,
    resolver_context,
)
from diwire._internal.providers import ProviderSpec
from diwire._internal.resolvers.assembly.compiler import ResolversAssemblyCompiler
from diwire.exceptions import (
    DIWireAsyncDependencyInSyncContextError,
    DIWireDependencyNotRegisteredError,
    DIWireScopeMismatchError,
)


class _SingletonService:
    pass


class _TransientService:
    pass


class _RequestService:
    pass


class _SessionService:
    pass


class _Resource:
    pass


class _DependsOnResource:
    def __init__(self, resource: _Resource) -> None:
        self.resource = resource


class _PosOnlyDependency:
    pass


class _PosOnlyService:
    def __init__(self, dependency: _PosOnlyDependency) -> None:
        self.dependency = dependency


class _VarArgsService:
    def __init__(self, values: tuple[int, ...]) -> None:
        self.values = values


class _KwArgsService:
    def __init__(self, options: dict[str, int]) -> None:
        self.options = options


class _SessionAsyncService:
    pass


class _InjectScopedDependency:
    pass


class _InjectScopedConsumer:
    def __init__(self, dependency: _InjectScopedDependency) -> None:
        self.dependency = dependency


class _InjectNestedInnerConsumer:
    def __init__(self, dependency: _InjectScopedDependency) -> None:
        self.dependency = dependency


class _InjectNestedOuterConsumer:
    def __init__(
        self,
        inner: _InjectNestedInnerConsumer,
        dependency: _InjectScopedDependency,
    ) -> None:
        self.inner = inner
        self.dependency = dependency


class _ContextValueConsumer:
    def __init__(self, value: int) -> None:
        self.value = value


class _RequestContextValue:
    def __init__(self, value: int) -> None:
        self.value = value


class _ActionContextValue:
    def __init__(self, value: int) -> None:
        self.value = value


class _StepContextValue:
    def __init__(self, value: int) -> None:
        self.value = value


class _MixedSharedSyncDependency:
    pass


class _MixedSyncConsumer:
    def __init__(self, dependency: _MixedSharedSyncDependency) -> None:
        self.dependency = dependency


class _MixedAsyncGraphDependency:
    pass


class _DataclassDefaultFactoryDependency:
    pass


@dataclass
class _DataclassDefaultFactoryConsumer:
    dependency: _DataclassDefaultFactoryDependency = field(
        default_factory=_DataclassDefaultFactoryDependency,
    )


class _DataclassDefaultValueDependency:
    pass


_DATACLASS_DEFAULT_VALUE_DEPENDENCY = _DataclassDefaultValueDependency()


@dataclass
class _DataclassDefaultValueConsumer:
    dependency: _DataclassDefaultValueDependency = _DATACLASS_DEFAULT_VALUE_DEPENDENCY


Q = TypeVar("Q")


class _OpenRuntimeService(Generic[Q]):
    pass


class _OpenRuntimeServiceImpl(_OpenRuntimeService[Q]):
    def __init__(self, type_arg: type[Q]) -> None:
        self.type_arg = type_arg


def _new_list_int_alias() -> Any:
    return list[int]


def _new_dict_str_int_alias() -> Any:
    return dict[str, int]


def _new_tuple_int_var_alias() -> Any:
    return tuple[int, ...]


def _new_int_or_str_alias() -> Any:
    return int | str


def _bound_self(method: Any) -> Any:
    return cast("Any", method).__self__


def test_sync_singleton_uses_lambda_method_replacement_cache() -> None:
    calls = 0

    def build_service() -> _SingletonService:
        nonlocal calls
        calls += 1
        return _SingletonService()

    container = Container()
    container.add_factory(
        build_service,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
    )

    first = container.resolve(_SingletonService)
    second = container.resolve(_SingletonService)

    assert first is second
    assert calls == 1

    root_resolver = container._root_resolver
    assert root_resolver is not None
    slot = container._providers_registrations.get_by_type(_SingletonService).slot
    cached_method = getattr(root_resolver, f"resolve_{slot}")
    assert callable(cached_method)
    assert cached_method.__name__ == "<lambda>"


@pytest.mark.asyncio
async def test_async_singleton_uses_async_cached_method_replacement() -> None:
    calls = 0

    async def build_service() -> _SingletonService:
        nonlocal calls
        calls += 1
        return _SingletonService()

    container = Container()
    container.add_factory(
        build_service,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
    )

    first = await container.aresolve(_SingletonService)
    second = await container.aresolve(_SingletonService)

    assert first is second
    assert calls == 1

    root_resolver = container._root_resolver
    assert root_resolver is not None
    slot = container._providers_registrations.get_by_type(_SingletonService).slot
    cached_method = getattr(root_resolver, f"aresolve_{slot}")
    assert inspect.iscoroutinefunction(cached_method)
    assert cached_method.__name__ == "_cached"


def test_compile_returns_cached_resolver_and_rebinds_entrypoints() -> None:
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
        use_resolver_context=False,
    )
    container.add_instance(_Resource(), provides=_Resource)

    initial_resolver = container._root_resolver
    assert initial_resolver is None
    assert _bound_self(container.resolve) is container
    assert _bound_self(container.aresolve) is container
    assert _bound_self(container.enter_scope) is container
    assert _bound_self(container.__enter__) is container
    assert _bound_self(container.__aenter__) is container
    assert _bound_self(container.__exit__) is container
    assert _bound_self(container.__aexit__) is container

    first_resolver = container.compile()
    second_resolver = container.compile()

    assert first_resolver is second_resolver
    assert container._root_resolver is first_resolver
    assert _bound_self(container.resolve) is first_resolver
    assert _bound_self(container.aresolve) is first_resolver
    assert _bound_self(container.enter_scope) is first_resolver
    assert _bound_self(container.__enter__) is container
    assert _bound_self(container.__aenter__) is container
    assert _bound_self(container.__exit__) is container
    assert _bound_self(container.__aexit__) is container


def test_resolve_auto_compiles_root_resolver_when_uncompiled() -> None:
    resource = _Resource()
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
        use_resolver_context=False,
    )
    container.add_instance(resource, provides=_Resource)

    initial_resolver = container._root_resolver
    assert initial_resolver is None

    resolved = container.resolve(_Resource)
    root_resolver = container._root_resolver

    assert resolved is resource
    assert root_resolver is not None
    assert _bound_self(container.resolve) is root_resolver


@pytest.mark.asyncio
async def test_aresolve_auto_compiles_root_resolver_when_uncompiled() -> None:
    resource = _Resource()
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
        use_resolver_context=False,
    )
    container.add_instance(resource, provides=_Resource)

    initial_resolver = container._root_resolver
    assert initial_resolver is None

    resolved = await container.aresolve(_Resource)
    root_resolver = container._root_resolver

    assert resolved is resource
    assert root_resolver is not None
    assert _bound_self(container.aresolve) is root_resolver


def test_compile_wraps_assembly_resolver_when_open_generic_registry_present() -> None:
    container = Container()
    container.add(_OpenRuntimeServiceImpl, provides=_OpenRuntimeService)

    compiled_resolver = container.compile()

    assert type(compiled_resolver).__name__ == "_ResolverBoundResolver"
    assert type(cast("Any", compiled_resolver)._resolver).__name__.endswith("OpenGenericResolver")
    assert _bound_self(container.resolve) is container
    resolved = container.resolve(_OpenRuntimeService[int])
    assert isinstance(resolved, _OpenRuntimeServiceImpl)
    assert resolved.type_arg is int


def test_enter_scope_auto_compiles_root_resolver_when_uncompiled() -> None:
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
        use_resolver_context=False,
    )
    container.add(
        _RequestService,
        provides=_RequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    initial_resolver = container._root_resolver
    assert initial_resolver is None

    with container.enter_scope() as request_scope:
        resolved = request_scope.resolve(_RequestService)
        assert isinstance(resolved, _RequestService)

    root_resolver = container._root_resolver
    assert root_resolver is not None
    assert _bound_self(container.enter_scope) is root_resolver


def test_registering_after_compile_invalidates_compilation_and_rebinds_lazy_entrypoints() -> None:
    class _RegisteredByInstance:
        pass

    class _RegisteredByConcrete:
        pass

    class _RegisteredByFactory:
        pass

    class _RegisteredByGenerator:
        pass

    class _RegisteredByContextManager:
        pass

    def build_factory() -> _RegisteredByFactory:
        return _RegisteredByFactory()

    def build_generator() -> Generator[_RegisteredByGenerator, None, None]:
        try:
            yield _RegisteredByGenerator()
        finally:
            pass

    class _ManagedContext:
        def __enter__(self) -> _RegisteredByContextManager:
            return _RegisteredByContextManager()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            return None

    def build_context_manager() -> _ManagedContext:
        return _ManagedContext()

    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
        use_resolver_context=False,
    )
    previous_resolver = container.compile()

    registrations: tuple[tuple[str, Any], ...] = (
        (
            "instance",
            lambda: container.add_instance(_RegisteredByInstance(), provides=_RegisteredByInstance),
        ),
        (
            "concrete",
            lambda: container.add(
                _RegisteredByConcrete,
                provides=_RegisteredByConcrete,
            ),
        ),
        (
            "factory",
            lambda: container.add_factory(
                build_factory,
                provides=_RegisteredByFactory,
            ),
        ),
        (
            "generator",
            lambda: container.add_generator(
                build_generator,
                provides=_RegisteredByGenerator,
            ),
        ),
        (
            "context manager",
            lambda: container.add_context_manager(
                build_context_manager,
                provides=_RegisteredByContextManager,
            ),
        ),
    )

    for registration_name, register in registrations:
        register()

        assert container._root_resolver is None, registration_name
        assert _bound_self(container.resolve) is container, registration_name
        assert _bound_self(container.aresolve) is container, registration_name
        assert _bound_self(container.enter_scope) is container, registration_name
        assert _bound_self(container.__enter__) is container, registration_name
        assert _bound_self(container.__aenter__) is container, registration_name
        assert _bound_self(container.__exit__) is container, registration_name
        assert _bound_self(container.__aexit__) is container, registration_name

        compiled_resolver = container.compile()
        assert compiled_resolver is not previous_resolver, registration_name
        assert _bound_self(container.resolve) is compiled_resolver, registration_name
        assert _bound_self(container.aresolve) is compiled_resolver, registration_name
        assert _bound_self(container.enter_scope) is compiled_resolver, registration_name
        assert _bound_self(container.__enter__) is container, registration_name
        assert _bound_self(container.__aenter__) is container, registration_name
        assert _bound_self(container.__exit__) is container, registration_name
        assert _bound_self(container.__aexit__) is container, registration_name
        previous_resolver = compiled_resolver


def test_autoregister_keeps_container_entrypoints_and_skips_existing_registration() -> None:
    class _AutoRegisteredService:
        pass

    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    first = container.resolve(_AutoRegisteredService)
    root_resolver = container._root_resolver

    assert isinstance(first, _AutoRegisteredService)
    assert root_resolver is not None
    assert _bound_self(container.resolve) is container
    assert _bound_self(container.aresolve) is container
    assert _bound_self(container.enter_scope) is container
    assert _bound_self(container.__enter__) is container
    assert _bound_self(container.__aenter__) is container
    assert _bound_self(container.__exit__) is container
    assert _bound_self(container.__aexit__) is container

    second = container.resolve(_AutoRegisteredService)

    assert isinstance(second, _AutoRegisteredService)
    assert _bound_self(container.resolve) is container
    assert _bound_self(container.aresolve) is container
    assert _bound_self(container.enter_scope) is container
    assert _bound_self(container.__enter__) is container
    assert _bound_self(container.__aenter__) is container
    assert _bound_self(container.__exit__) is container
    assert _bound_self(container.__aexit__) is container
    assert container._root_resolver is root_resolver


def test_dunder_enter_and_exit_delegate_to_root_resolver() -> None:
    container = Container()
    container.add_instance(_Resource(), provides=_Resource)

    entered = container.__enter__()
    root_resolver = container._root_resolver

    assert root_resolver is not None
    assert entered is root_resolver

    container.__exit__(None, None, None)


def test_dunder_exit_without_matching_enter_raises_runtime_error() -> None:
    container = Container()

    with pytest.raises(RuntimeError, match="without a matching enter"):
        container.__exit__(None, None, None)

    with pytest.raises(RuntimeError, match="without a matching enter"):
        container.close()


def test_container_exit_wrapper_delegates_when_root_resolver_exists() -> None:
    container = Container()
    container.add_instance(_Resource(), provides=_Resource)
    container.compile()

    Container.__exit__(container, None, None, None)


@pytest.mark.asyncio
async def test_dunder_aenter_and_aexit_delegate_to_root_resolver() -> None:
    container = Container()
    container.add_instance(_Resource(), provides=_Resource)

    entered = await container.__aenter__()
    root_resolver = container._root_resolver

    assert root_resolver is not None
    assert entered is root_resolver

    await container.__aexit__(None, None, None)


@pytest.mark.asyncio
async def test_dunder_aexit_without_matching_aenter_raises_runtime_error() -> None:
    container = Container()

    with pytest.raises(RuntimeError, match="without a matching enter"):
        await container.__aexit__(None, None, None)

    with pytest.raises(RuntimeError, match="without a matching enter"):
        await container.aclose()


@pytest.mark.asyncio
async def test_container_aexit_wrapper_delegates_when_root_resolver_exists() -> None:
    container = Container()
    container.add_instance(_Resource(), provides=_Resource)
    container.compile()

    await Container.__aexit__(container, None, None, None)


def test_transient_provider_is_not_cached() -> None:
    calls = 0

    def build_service() -> _TransientService:
        nonlocal calls
        calls += 1
        return _TransientService()

    container = Container()
    container.add_factory(
        build_service,
        provides=_TransientService,
        lifetime=Lifetime.TRANSIENT,
    )

    first = container.resolve(_TransientService)
    second = container.resolve(_TransientService)

    assert first is not second
    assert calls == 2


def test_scope_resolution_requires_explicit_opened_scope_chain() -> None:
    container = Container()
    container.add(
        _RequestService,
        provides=_RequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(DIWireScopeMismatchError):
        container.resolve(_RequestService)

    with container.enter_scope(Scope.ACTION) as action_scope:
        from_action = action_scope.resolve(_RequestService)
        assert isinstance(from_action, _RequestService)
        assert from_action is action_scope.resolve(_RequestService)

    with container.enter_scope() as request_scope:
        with request_scope.enter_scope(Scope.ACTION) as action_scope:
            from_request = request_scope.resolve(_RequestService)
            from_action = action_scope.resolve(_RequestService)
            assert from_request is from_action


def test_contextvar_dependency_is_resolved_and_token_reset_restores_default() -> None:
    current_value_var: ContextVar[int] = ContextVar("runtime_context_value", default=0)

    def read_current_value() -> int:
        return current_value_var.get()

    def build_consumer(value: int) -> _ContextValueConsumer:
        return _ContextValueConsumer(value=value)

    container = Container()
    container.add_factory(
        read_current_value,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        build_consumer,
        provides=_ContextValueConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        assert request_scope.resolve(_ContextValueConsumer).value == 0
        token = current_value_var.set(123)
        try:
            assert request_scope.resolve(_ContextValueConsumer).value == 123
        finally:
            current_value_var.reset(token)
        assert request_scope.resolve(_ContextValueConsumer).value == 0


def test_contextvar_visibility_includes_scope_and_children_with_nearest_override() -> None:
    current_value_var: ContextVar[int] = ContextVar("runtime_nested_value", default=1)

    def read_current_value() -> int:
        return current_value_var.get()

    def build_request_value(value: int) -> _RequestContextValue:
        return _RequestContextValue(value=value)

    def build_action_value(value: int) -> _ActionContextValue:
        return _ActionContextValue(value=value)

    def build_step_value(value: int) -> _StepContextValue:
        return _StepContextValue(value=value)

    container = Container()
    container.add_factory(
        read_current_value,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        build_request_value,
        provides=_RequestContextValue,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        build_action_value,
        provides=_ActionContextValue,
        scope=Scope.ACTION,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        build_step_value,
        provides=_StepContextValue,
        scope=Scope.STEP,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        assert request_scope.resolve(_RequestContextValue).value == 1
        with request_scope.enter_scope(Scope.ACTION) as action_scope:
            assert action_scope.resolve(_ActionContextValue).value == 1
            token = current_value_var.set(2)
            try:
                with action_scope.enter_scope(Scope.STEP) as step_scope:
                    assert step_scope.resolve(_StepContextValue).value == 2
                    assert step_scope.resolve(_RequestContextValue).value == 2
            finally:
                current_value_var.reset(token)
            assert action_scope.resolve(_ActionContextValue).value == 1


def test_scope_resolver_can_directly_resolve_contextvar_backed_dependency() -> None:
    current_value_var: ContextVar[int] = ContextVar("runtime_direct_value", default=0)
    container = Container()

    def read_current_value() -> int:
        return current_value_var.get()

    container.add_factory(
        read_current_value,
        provides=int,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        token = current_value_var.set(9)
        try:
            assert request_scope.resolve(int) == 9
        finally:
            current_value_var.reset(token)


def test_container_resolve_normalizes_non_component_metadata_for_regular_dependencies() -> None:
    priority_key = Annotated[int, Component("priority")]
    container = Container()
    container.add_instance(9, provides=int)
    container.add_instance(7, provides=priority_key)

    assert container.resolve(Annotated[int, "meta"]) == 9
    assert container.resolve(Annotated[int, Component("priority"), "meta"]) == 7


def test_assembly_passes_resolver_to_inject_wrapped_provider_calls() -> None:
    container = Container()
    container.add(
        _InjectScopedDependency,
        provides=_InjectScopedDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @resolver_context.inject
    def build_consumer(
        dependency: Injected[_InjectScopedDependency],
    ) -> _InjectScopedConsumer:
        return _InjectScopedConsumer(dependency=dependency)

    container.add_factory(
        build_consumer,
        provides=_InjectScopedConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        resolved = request_scope.resolve(_InjectScopedConsumer)
        assert isinstance(resolved, _InjectScopedConsumer)
        assert isinstance(resolved.dependency, _InjectScopedDependency)


@pytest.mark.asyncio
async def test_assembly_async_inject_wrapper_provider_receives_resolver() -> None:
    container = Container()
    container.add(
        _InjectScopedDependency,
        provides=_InjectScopedDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @resolver_context.inject
    async def build_consumer(
        dependency: Injected[_InjectScopedDependency],
    ) -> _InjectScopedConsumer:
        await asyncio.sleep(0)
        return _InjectScopedConsumer(dependency=dependency)

    container.add_factory(
        build_consumer,
        provides=_InjectScopedConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        resolved = await request_scope.aresolve(_InjectScopedConsumer)

    assert isinstance(resolved, _InjectScopedConsumer)
    assert isinstance(resolved.dependency, _InjectScopedDependency)


def test_assembly_nested_inject_wrappers_runtime_scope_consistency() -> None:
    container = Container()
    container.add(
        _InjectScopedDependency,
        provides=_InjectScopedDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    @resolver_context.inject
    def build_inner(
        dependency: Injected[_InjectScopedDependency],
    ) -> _InjectNestedInnerConsumer:
        return _InjectNestedInnerConsumer(dependency=dependency)

    @resolver_context.inject
    def build_outer(
        inner: Injected[_InjectNestedInnerConsumer],
        dependency: Injected[_InjectScopedDependency],
    ) -> _InjectNestedOuterConsumer:
        return _InjectNestedOuterConsumer(inner=inner, dependency=dependency)

    container.add_factory(
        build_inner,
        provides=_InjectNestedInnerConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_factory(
        build_outer,
        provides=_InjectNestedOuterConsumer,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope() as request_scope:
        resolved = request_scope.resolve(_InjectNestedOuterConsumer)

    assert isinstance(resolved, _InjectNestedOuterConsumer)
    assert resolved.inner.dependency is resolved.dependency


def test_assembly_inject_wrapper_singleton_thread_safe_stress() -> None:
    calls = 0
    workers = 32

    dependency = _Resource()
    container = Container()
    container.add_instance(dependency, provides=_Resource)

    @resolver_context.inject
    def build_singleton(resource: Injected[_Resource]) -> _DependsOnResource:
        nonlocal calls
        calls += 1
        return _DependsOnResource(resource=resource)

    container.add_factory(
        build_singleton,
        provides=_DependsOnResource,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.THREAD,
    )
    container.compile()

    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = [pool.submit(container.resolve, _DependsOnResource) for _ in range(workers * 2)]
        results = [future.result(timeout=5) for future in futures]

    assert calls == 1
    assert len({id(result) for result in results}) == 1


@pytest.mark.asyncio
async def test_assembly_inject_wrapper_singleton_async_stress() -> None:
    calls = 0
    tasks = 128
    dependency = _Resource()
    container = Container()
    container.add_instance(dependency, provides=_Resource)

    @resolver_context.inject
    async def build_singleton(resource: Injected[_Resource]) -> _DependsOnResource:
        nonlocal calls
        calls += 1
        await asyncio.sleep(0)
        return _DependsOnResource(resource=resource)

    container.add_factory(
        build_singleton,
        provides=_DependsOnResource,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.ASYNC,
    )

    results = await asyncio.gather(
        *(container.aresolve(_DependsOnResource) for _ in range(tasks)),
    )

    assert calls == 1
    assert len({id(result) for result in results}) == 1


def test_assembly_inject_wrapper_unsafe_mode_stress_no_deadlock() -> None:
    calls = 0
    workers = 32
    all_started = threading.Event()
    calls_lock = threading.Lock()
    dependency = _Resource()
    container = Container()
    container.add_instance(dependency, provides=_Resource)

    @resolver_context.inject
    def build_singleton(resource: Injected[_Resource]) -> _DependsOnResource:
        nonlocal calls
        with calls_lock:
            calls += 1
            if calls == workers:
                all_started.set()
        did_start = all_started.wait(timeout=2)
        assert did_start
        return _DependsOnResource(resource=resource)

    container.add_factory(
        build_singleton,
        provides=_DependsOnResource,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.NONE,
    )
    container.compile()

    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = [pool.submit(container.resolve, _DependsOnResource) for _ in range(workers)]
        results = [future.result(timeout=5) for future in futures]

    assert calls == workers
    assert all(isinstance(result, _DependsOnResource) for result in results)
    assert len({id(result) for result in results}) == workers
    assert container.resolve(_DependsOnResource) is container.resolve(_DependsOnResource)


def test_generated_dispatch_raises_for_unknown_dependency_in_sync_and_async_paths() -> None:
    container = Container()

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(object)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        asyncio.run(container.aresolve(object))

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(_new_list_int_alias())

    with pytest.raises(DIWireDependencyNotRegisteredError):
        asyncio.run(container.aresolve(_new_list_int_alias()))


def test_missing_dataclass_default_factory_dependency_uses_default_factory() -> None:
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )
    container.add(
        _DataclassDefaultFactoryConsumer,
        provides=_DataclassDefaultFactoryConsumer,
    )

    resolved = container.resolve(_DataclassDefaultFactoryConsumer)

    assert isinstance(resolved.dependency, _DataclassDefaultFactoryDependency)


def test_registered_dataclass_default_factory_dependency_overrides_default_factory() -> None:
    dependency = _DataclassDefaultFactoryDependency()
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )
    container.add_instance(dependency, provides=_DataclassDefaultFactoryDependency)
    container.add(
        _DataclassDefaultFactoryConsumer,
        provides=_DataclassDefaultFactoryConsumer,
    )

    resolved = container.resolve(_DataclassDefaultFactoryConsumer)

    assert resolved.dependency is dependency


def test_missing_dataclass_default_value_dependency_still_errors() -> None:
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )
    container.add(
        _DataclassDefaultValueConsumer,
        provides=_DataclassDefaultValueConsumer,
    )

    with pytest.raises(DIWireDependencyNotRegisteredError, match="is not registered"):
        container.resolve(_DataclassDefaultValueConsumer)


@pytest.mark.parametrize(
    "alias_factory",
    [
        _new_list_int_alias,
        _new_dict_str_int_alias,
        _new_tuple_int_var_alias,
        _new_int_or_str_alias,
    ],
)
def test_generated_dispatch_resolves_equal_non_identical_generic_alias_keys(
    alias_factory: Callable[[], Any],
) -> None:
    registration_key = alias_factory()
    lookup_key = alias_factory()
    instance = object()
    container = Container()
    container.add_instance(instance, provides=registration_key)

    assert lookup_key == registration_key
    assert lookup_key is not registration_key
    assert container.resolve(lookup_key) is instance


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "alias_factory",
    [
        _new_list_int_alias,
        _new_dict_str_int_alias,
        _new_tuple_int_var_alias,
        _new_int_or_str_alias,
    ],
)
async def test_generated_aresolve_resolves_equal_non_identical_generic_alias_keys(
    alias_factory: Callable[[], Any],
) -> None:
    registration_key = alias_factory()
    lookup_key = alias_factory()
    instance = object()
    container = Container()
    container.add_instance(instance, provides=registration_key)

    assert lookup_key == registration_key
    assert lookup_key is not registration_key
    assert await container.aresolve(lookup_key) is instance


def test_runtime_compiler_handles_unusual_module_and_qualname_symbols() -> None:
    bad_symbol = type('Bad"""Name', (), {"__module__": 'bad"""module'})
    bad_instance = bad_symbol()
    container = Container()
    container.add_instance(bad_instance, provides=bad_symbol)
    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )
    assert root_resolver.resolve(bad_symbol) is bad_instance


def test_enter_scope_can_choose_skippable_or_next_non_skippable_scope() -> None:
    container = Container()
    container.add(
        _SessionService,
        provides=_SessionService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _RequestService,
        provides=_RequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.SESSION) as session_scope:
        session_instance = session_scope.resolve(_SessionService)
        with session_scope.enter_scope() as request_scope:
            request_instance = request_scope.resolve(_RequestService)
            assert request_instance is request_scope.resolve(_RequestService)
        assert session_instance is session_scope.resolve(_SessionService)

    with container.enter_scope() as request_scope:
        assert isinstance(request_scope.resolve(_RequestService), _RequestService)
        with pytest.raises(DIWireScopeMismatchError):
            request_scope.resolve(_SessionService)


def test_enter_scope_returns_self_for_same_scope_and_rejects_non_forward_transition() -> None:
    container = Container()
    with container.enter_scope(Scope.SESSION) as session_scope:
        assert session_scope.enter_scope(Scope.SESSION) is session_scope

        with session_scope.enter_scope() as request_scope:
            assert request_scope.enter_scope(Scope.REQUEST) is request_scope
            with pytest.raises(DIWireScopeMismatchError, match="Cannot enter scope level"):
                request_scope.enter_scope(Scope.SESSION)


def test_enter_scope_from_deepest_scope_without_target_raises() -> None:
    container = Container()
    with container.enter_scope(Scope.SESSION) as session_scope:
        with session_scope.enter_scope() as request_scope:
            with request_scope.enter_scope() as action_scope:
                with action_scope.enter_scope() as step_scope:
                    with pytest.raises(DIWireScopeMismatchError, match="Cannot enter deeper scope"):
                        step_scope.enter_scope()


def test_request_resolver_delegates_to_session_owner_cache_when_opened_from_session() -> None:
    calls = 0

    def build_session_service() -> _SessionService:
        nonlocal calls
        calls += 1
        return _SessionService()

    container = Container()
    container.add_factory(
        build_session_service,
        provides=_SessionService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.SESSION) as session_scope:
        session_instance = session_scope.resolve(_SessionService)
        with session_scope.enter_scope() as request_scope:
            delegated_instance = request_scope.resolve(_SessionService)
            delegated_again = request_scope.resolve(_SessionService)

    assert session_instance is delegated_instance
    assert delegated_instance is delegated_again
    assert calls == 1


def test_request_resolve_for_sync_session_service_raises_when_owner_missing() -> None:
    container = Container()
    container.add_factory(
        _SessionService,
        provides=_SessionService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        with pytest.raises(DIWireScopeMismatchError, match="requires opened scope level"):
            request_scope.resolve(_SessionService)


@pytest.mark.asyncio
async def test_request_aresolve_for_async_session_service_raises_when_owner_missing() -> None:
    async def build_session_async_service() -> _SessionAsyncService:
        return _SessionAsyncService()

    container = Container()
    container.add_factory(
        build_session_async_service,
        provides=_SessionAsyncService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        with pytest.raises(DIWireScopeMismatchError, match="requires opened scope level"):
            await request_scope.aresolve(_SessionAsyncService)


@pytest.mark.asyncio
async def test_request_resolver_async_delegates_to_session_owner_cache() -> None:
    calls = 0

    async def build_session_async_service() -> _SessionAsyncService:
        nonlocal calls
        calls += 1
        return _SessionAsyncService()

    container = Container()
    container.add_factory(
        build_session_async_service,
        provides=_SessionAsyncService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.SESSION) as session_scope:
        session_instance = await session_scope.aresolve(_SessionAsyncService)
        with session_scope.enter_scope() as request_scope:
            delegated_instance = await request_scope.aresolve(_SessionAsyncService)
            delegated_again = await request_scope.aresolve(_SessionAsyncService)

    assert session_instance is delegated_instance
    assert delegated_instance is delegated_again
    assert calls == 1


def test_runtime_compiler_uses_thread_locks_for_sync_graph() -> None:
    container = Container(lock_mode="auto")
    container.add_factory(
        _SingletonService,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
    )

    slot = container._providers_registrations.get_by_type(_SingletonService).slot
    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )
    runtime = cast("Any", type(root_resolver))._runtime

    assert slot in runtime.thread_lock_by_slot
    assert runtime.async_lock_by_slot == {}


@pytest.mark.asyncio
async def test_renderer_emits_async_locks_only_for_async_cached_paths() -> None:
    async def build_async() -> _SingletonService:
        return _SingletonService()

    container = Container(lock_mode="auto")
    container.add_factory(
        build_async,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
    )
    container.add_factory(
        _TransientService,
        provides=_TransientService,
        lifetime=Lifetime.SCOPED,
    )

    async_slot = container._providers_registrations.get_by_type(_SingletonService).slot
    sync_slot = container._providers_registrations.get_by_type(_TransientService).slot

    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )
    runtime = cast("Any", type(root_resolver))._runtime

    assert async_slot in runtime.async_lock_by_slot
    assert sync_slot not in runtime.async_lock_by_slot
    assert runtime.thread_lock_by_slot == {}

    resolved = await container.aresolve(_TransientService)
    assert isinstance(resolved, _TransientService)


def test_runtime_compiler_lock_mode_none_uses_no_locks() -> None:
    container = Container()
    container.add_factory(
        _SingletonService,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.NONE,
    )

    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )
    runtime = cast("Any", type(root_resolver))._runtime

    assert runtime.thread_lock_by_slot == {}
    assert runtime.async_lock_by_slot == {}


def test_mixed_graph_thread_override_keeps_sync_singleton_thread_safe() -> None:
    calls = 0
    workers = 24

    def build_sync_singleton() -> _SingletonService:
        nonlocal calls
        calls += 1
        return _SingletonService()

    async def build_async_singleton() -> _TransientService:
        return _TransientService()

    container = Container()
    container.add_factory(
        build_sync_singleton,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.THREAD,
    )
    container.add_factory(
        build_async_singleton,
        provides=_TransientService,
        lifetime=Lifetime.SCOPED,
    )
    container.compile()

    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = [pool.submit(container.resolve, _SingletonService) for _ in range(workers * 2)]
        results = [future.result(timeout=5) for future in futures]

    assert calls == 1
    assert len({id(result) for result in results}) == 1


def test_mixed_graph_thread_override_keeps_sync_dependency_chain_thread_safe() -> None:
    shared_calls = 0
    consumer_calls = 0
    workers = 24

    def build_shared_dependency() -> _MixedSharedSyncDependency:
        nonlocal shared_calls
        shared_calls += 1
        return _MixedSharedSyncDependency()

    def build_consumer(dependency: _MixedSharedSyncDependency) -> _MixedSyncConsumer:
        nonlocal consumer_calls
        consumer_calls += 1
        return _MixedSyncConsumer(dependency=dependency)

    async def build_async_graph_dependency() -> _MixedAsyncGraphDependency:
        return _MixedAsyncGraphDependency()

    container = Container(lock_mode="auto")
    container.add_factory(
        build_shared_dependency,
        provides=_MixedSharedSyncDependency,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.THREAD,
    )
    container.add_factory(
        build_consumer,
        provides=_MixedSyncConsumer,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.THREAD,
    )
    container.add_factory(
        build_async_graph_dependency,
        provides=_MixedAsyncGraphDependency,
        lifetime=Lifetime.SCOPED,
    )

    shared_slot = container._providers_registrations.get_by_type(_MixedSharedSyncDependency).slot
    consumer_slot = container._providers_registrations.get_by_type(_MixedSyncConsumer).slot
    async_slot = container._providers_registrations.get_by_type(_MixedAsyncGraphDependency).slot
    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )
    runtime = cast("Any", type(root_resolver))._runtime

    assert shared_slot in runtime.thread_lock_by_slot
    assert consumer_slot in runtime.thread_lock_by_slot
    assert async_slot in runtime.async_lock_by_slot

    container._root_resolver = root_resolver

    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = [pool.submit(container.resolve, _MixedSyncConsumer) for _ in range(workers * 2)]
        results = [future.result(timeout=5) for future in futures]

    assert shared_calls == 1
    assert consumer_calls == 1
    assert len({id(result) for result in results}) == 1
    assert len({id(result.dependency) for result in results}) == 1


def test_lock_mode_async_on_sync_cached_provider_leaves_sync_path_unlocked() -> None:
    calls = 0
    workers = 12
    all_started = threading.Event()
    calls_lock = threading.Lock()

    def build_singleton() -> _SingletonService:
        nonlocal calls
        with calls_lock:
            calls += 1
            if calls == workers:
                all_started.set()
        did_start = all_started.wait(timeout=2)
        assert did_start
        return _SingletonService()

    container = Container()
    container.add_factory(
        build_singleton,
        provides=_SingletonService,
        lifetime=Lifetime.SCOPED,
        lock_mode=LockMode.ASYNC,
    )
    slot = container._providers_registrations.get_by_type(_SingletonService).slot
    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
    )
    runtime = cast("Any", type(root_resolver))._runtime
    assert slot not in runtime.thread_lock_by_slot
    container._root_resolver = root_resolver

    with ThreadPoolExecutor(max_workers=workers) as pool:
        futures = [pool.submit(container.resolve, _SingletonService) for _ in range(workers)]
        results = [future.result(timeout=5) for future in futures]

    assert calls == workers
    assert len({id(result) for result in results}) == workers


def test_sync_generator_runs_cleanup_on_scope_exit() -> None:
    events: list[str] = []

    def provide_resource() -> Generator[_Resource, None, None]:
        events.append("enter")
        try:
            yield _Resource()
        finally:
            events.append("exit")

    container = Container()
    container.add_generator(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        resolved = request_scope.resolve(_Resource)
        assert isinstance(resolved, _Resource)
        assert events == ["enter"]

    assert events == ["enter", "exit"]


def test_sync_generator_runs_cleanup_for_implicit_request_scope_on_action_exit() -> None:
    events: list[str] = []

    def provide_resource() -> Generator[_Resource, None, None]:
        events.append("enter")
        try:
            yield _Resource()
        finally:
            events.append("exit")

    container = Container()
    container.add_generator(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.ACTION) as action_scope:
        resolved = action_scope.resolve(_Resource)
        assert isinstance(resolved, _Resource)
        assert events == ["enter"]

    assert events == ["enter", "exit"]


def test_sync_generator_cleanup_disabled_keeps_cleanup_callbacks_empty() -> None:
    events: list[str] = []

    def provide_resource() -> Generator[_Resource, None, None]:
        events.append("enter")
        try:
            yield _Resource()
        finally:
            pass

    container = Container()
    container.add_generator(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
        cleanup_enabled=False,
    )

    request_scope = root_resolver.enter_scope()
    resolved = request_scope.resolve(_Resource)
    request_scope_any = cast("Any", request_scope)

    assert isinstance(resolved, _Resource)
    assert request_scope_any._cleanup_callbacks == []
    request_scope.__exit__(None, None, None)
    assert request_scope_any._cleanup_callbacks == []
    assert events == ["enter"]


@pytest.mark.asyncio
async def test_async_generator_runs_cleanup_on_async_scope_exit() -> None:
    events: list[str] = []

    async def provide_resource() -> AsyncGenerator[_Resource, None]:
        events.append("enter")
        try:
            yield _Resource()
        finally:
            events.append("exit")

    container = Container()
    container.add_generator(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    async with container.enter_scope() as request_scope:
        resolved = await request_scope.aresolve(_Resource)
        assert isinstance(resolved, _Resource)
        assert events == ["enter"]

    assert events == ["enter", "exit"]


def test_sync_resolution_for_async_generator_raises_async_dependency_error() -> None:
    async def provide_resource() -> AsyncGenerator[_Resource, None]:
        try:
            yield _Resource()
        finally:
            pass

    container = Container()
    container.add_generator(provide_resource, provides=_Resource)

    with pytest.raises(
        DIWireAsyncDependencyInSyncContextError,
        match="requires asynchronous resolution",
    ):
        container.resolve(_Resource)


@pytest.mark.asyncio
async def test_async_context_manager_provider_uses_aenter_and_aexit() -> None:
    enter_calls = 0
    exit_calls = 0

    class _ManagedAsyncContext:
        async def __aenter__(self) -> _Resource:
            nonlocal enter_calls
            enter_calls += 1
            return _Resource()

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            nonlocal exit_calls
            exit_calls += 1

    def provide_resource() -> AbstractAsyncContextManager[_Resource]:
        return _ManagedAsyncContext()

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    async with container.enter_scope() as request_scope:
        resolved = await request_scope.aresolve(_Resource)
        assert isinstance(resolved, _Resource)

    assert enter_calls == 1
    assert exit_calls == 1


def test_sync_scope_exit_raises_when_async_cleanup_callbacks_are_present() -> None:
    async def provide_resource() -> AsyncGenerator[_Resource, None]:
        try:
            yield _Resource()
        finally:
            pass

    container = Container()
    container.add_generator(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(
        DIWireAsyncDependencyInSyncContextError,
        match="Cannot execute async cleanup in sync context",
    ):
        with container.enter_scope() as request_scope:
            asyncio.run(request_scope.aresolve(_Resource))


def test_cleanup_callbacks_execute_in_lifo_order() -> None:
    events: list[str] = []

    class _FirstContext:
        def __enter__(self) -> _Resource:
            events.append("first-enter")
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            events.append("first-exit")

    class _SecondContext:
        def __enter__(self) -> _SingletonService:
            events.append("second-enter")
            return _SingletonService()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            events.append("second-exit")

    def provide_first() -> _FirstContext:
        return _FirstContext()

    def provide_second() -> _SecondContext:
        return _SecondContext()

    class _Both:
        def __init__(self, first: _Resource, second: _SingletonService) -> None:
            self.first = first
            self.second = second

    container = Container()
    container.add_context_manager(
        provide_first,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(
        provide_second,
        provides=_SingletonService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _Both,
        provides=_Both,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        request_scope.resolve(_Both)

    assert events == ["first-enter", "second-enter", "second-exit", "first-exit"]


def test_resolver_close_matches_with_cleanup_order_for_generator_and_context_manager() -> None:
    events: list[str] = []

    def provide_generator() -> Generator[_Resource, None, None]:
        events.append("generator-enter")
        try:
            yield _Resource()
        finally:
            events.append("generator-exit")

    class _TrackedContext:
        def __enter__(self) -> _SingletonService:
            events.append("context-enter")
            return _SingletonService()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            events.append("context-exit")

    def provide_context() -> _TrackedContext:
        return _TrackedContext()

    class _Both:
        def __init__(self, first: _Resource, second: _SingletonService) -> None:
            self.first = first
            self.second = second

    container = Container()
    container.add_generator(
        provide_generator,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(
        provide_context,
        provides=_SingletonService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _Both,
        provides=_Both,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        request_scope.resolve(_Both)

    with_events = list(events)
    events.clear()

    request_scope = container.enter_scope()
    request_scope.resolve(_Both)
    request_scope.close()

    assert events == with_events
    assert events == ["generator-enter", "context-enter", "context-exit", "generator-exit"]


def test_cleanup_keeps_first_cleanup_error_when_multiple_callbacks_fail() -> None:
    class _FirstContext:
        def __enter__(self) -> _Resource:
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            raise RuntimeError("first-exit-error")

    class _SecondContext:
        def __enter__(self) -> _SingletonService:
            return _SingletonService()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            raise RuntimeError("second-exit-error")

    def provide_first() -> _FirstContext:
        return _FirstContext()

    def provide_second() -> _SecondContext:
        return _SecondContext()

    class _Both:
        def __init__(self, first: _Resource, second: _SingletonService) -> None:
            self.first = first
            self.second = second

    container = Container()
    container.add_context_manager(
        provide_first,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(
        provide_second,
        provides=_SingletonService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _Both,
        provides=_Both,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(RuntimeError, match="second-exit-error"):
        with container.enter_scope() as request_scope:
            request_scope.resolve(_Both)


def test_cleanup_does_not_override_active_body_exception() -> None:
    class _FailingContext:
        def __enter__(self) -> _Resource:
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            raise RuntimeError("cleanup boom")

    def provide_resource() -> _FailingContext:
        return _FailingContext()

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(ValueError, match="body boom"):
        with container.enter_scope() as request_scope:
            request_scope.resolve(_Resource)
            raise ValueError("body boom")


@pytest.mark.asyncio
async def test_async_cleanup_callbacks_execute_in_lifo_order_and_drain_stack() -> None:
    events: list[str] = []

    class _SyncContext:
        def __enter__(self) -> _Resource:
            events.append("sync-enter")
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            events.append("sync-exit")

    class _AsyncContext:
        async def __aenter__(self) -> _SingletonService:
            events.append("async-enter")
            return _SingletonService()

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            events.append("async-exit")

    def provide_sync() -> _SyncContext:
        return _SyncContext()

    def provide_async() -> AbstractAsyncContextManager[_SingletonService]:
        return _AsyncContext()

    class _Both:
        def __init__(self, first: _Resource, second: _SingletonService) -> None:
            self.first = first
            self.second = second

    container = Container()
    container.add_context_manager(
        provide_sync,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(
        provide_async,
        provides=_SingletonService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _Both,
        provides=_Both,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    request_scope: Any = None
    async with container.enter_scope() as opened_scope:
        request_scope = opened_scope
        await opened_scope.aresolve(_Both)

    assert events == ["sync-enter", "async-enter", "async-exit", "sync-exit"]
    assert request_scope._cleanup_callbacks == []


@pytest.mark.asyncio
async def test_resolver_aclose_matches_async_with_cleanup_order_for_async_providers() -> None:
    events: list[str] = []

    async def provide_generator() -> AsyncGenerator[_Resource, None]:
        events.append("generator-enter")
        try:
            yield _Resource()
        finally:
            events.append("generator-exit")

    @asynccontextmanager
    async def provide_context() -> AsyncGenerator[_SingletonService, None]:
        events.append("context-enter")
        try:
            yield _SingletonService()
        finally:
            events.append("context-exit")

    class _Both:
        def __init__(self, first: _Resource, second: _SingletonService) -> None:
            self.first = first
            self.second = second

    container = Container()
    container.add_generator(
        provide_generator,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(
        provide_context,
        provides=_SingletonService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _Both,
        provides=_Both,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    async with container.enter_scope() as request_scope:
        await request_scope.aresolve(_Both)

    async_with_events = list(events)
    events.clear()

    request_scope = container.enter_scope()
    await request_scope.aresolve(_Both)
    await request_scope.aclose()

    assert events == async_with_events
    assert events == ["generator-enter", "context-enter", "context-exit", "generator-exit"]


@pytest.mark.asyncio
async def test_async_cleanup_keeps_first_cleanup_error_when_multiple_callbacks_fail() -> None:
    class _FirstAsyncContext:
        async def __aenter__(self) -> _Resource:
            return _Resource()

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            raise RuntimeError("first-async-exit-error")

    class _SecondAsyncContext:
        async def __aenter__(self) -> _SingletonService:
            return _SingletonService()

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            raise RuntimeError("second-async-exit-error")

    def provide_first() -> AbstractAsyncContextManager[_Resource]:
        return _FirstAsyncContext()

    def provide_second() -> AbstractAsyncContextManager[_SingletonService]:
        return _SecondAsyncContext()

    class _Both:
        def __init__(self, first: _Resource, second: _SingletonService) -> None:
            self.first = first
            self.second = second

    container = Container()
    container.add_context_manager(
        provide_first,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(
        provide_second,
        provides=_SingletonService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _Both,
        provides=_Both,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(RuntimeError, match="second-async-exit-error"):
        async with container.enter_scope() as request_scope:
            await request_scope.aresolve(_Both)


@pytest.mark.asyncio
async def test_async_cleanup_does_not_override_body_exception_and_forwards_exc_info() -> None:
    received: dict[str, object | None] = {
        "exc_type": None,
        "exc_value": None,
    }

    class _TrackingAsyncContext:
        async def __aenter__(self) -> _Resource:
            return _Resource()

        async def __aexit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            received["exc_type"] = exc_type
            received["exc_value"] = exc_value
            raise RuntimeError("async cleanup boom")

    def provide_resource() -> AbstractAsyncContextManager[_Resource]:
        return _TrackingAsyncContext()

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(ValueError, match="async body boom"):
        async with container.enter_scope() as request_scope:
            await request_scope.aresolve(_Resource)
            raise ValueError("async body boom")

    assert received["exc_type"] is ValueError
    assert isinstance(received["exc_value"], ValueError)


@pytest.mark.asyncio
async def test_async_generator_dependency_chain_requires_aresolve() -> None:
    async def provide_resource() -> AsyncGenerator[_Resource, None]:
        try:
            yield _Resource()
        finally:
            pass

    def build_dependent(resource: _Resource) -> _DependsOnResource:
        return _DependsOnResource(resource)

    container = Container()
    container.add_generator(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_factory(
        build_dependent,
        provides=_DependsOnResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        with pytest.raises(
            DIWireAsyncDependencyInSyncContextError,
            match="requires asynchronous resolution",
        ):
            request_scope.resolve(_DependsOnResource)

    async with container.enter_scope() as request_scope:
        resolved = await request_scope.aresolve(_DependsOnResource)
        assert isinstance(resolved, _DependsOnResource)
        assert isinstance(resolved.resource, _Resource)


@pytest.mark.asyncio
async def test_async_context_manager_dependency_chain_requires_aresolve() -> None:
    @asynccontextmanager
    async def provide_resource() -> AsyncGenerator[_Resource, None]:
        yield _Resource()

    def build_dependent(resource: _Resource) -> _DependsOnResource:
        return _DependsOnResource(resource)

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_factory(
        build_dependent,
        provides=_DependsOnResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        with pytest.raises(
            DIWireAsyncDependencyInSyncContextError,
            match="requires asynchronous resolution",
        ):
            request_scope.resolve(_DependsOnResource)

    async with container.enter_scope() as request_scope:
        resolved = await request_scope.aresolve(_DependsOnResource)
        assert isinstance(resolved, _DependsOnResource)
        assert isinstance(resolved.resource, _Resource)


@pytest.mark.asyncio
async def test_async_scope_mismatch_raises_for_deeper_scoped_provider() -> None:
    container = Container()
    container.add(
        _RequestService,
        provides=_RequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(DIWireScopeMismatchError, match="requires opened scope level"):
        await container.aresolve(_RequestService)


def test_positional_only_dependency_wiring_is_supported() -> None:
    dependency_instance = _PosOnlyDependency()

    def build_service(dependency: _PosOnlyDependency, /) -> _PosOnlyService:
        return _PosOnlyService(dependency)

    container = Container()
    container.add_instance(dependency_instance, provides=_PosOnlyDependency)
    container.add_factory(build_service, provides=_PosOnlyService)

    resolved = container.resolve(_PosOnlyService)

    assert isinstance(resolved, _PosOnlyService)
    assert resolved.dependency is dependency_instance


def test_var_positional_dependency_wiring_expands_iterable_dependency() -> None:
    def build_service(*values: int) -> _VarArgsService:
        return _VarArgsService(values)

    signature = inspect.signature(build_service)
    values_parameter = signature.parameters["values"]
    values_type = tuple[int, ...]
    values_instance = (1, 2, 3)

    container = Container()
    container.add_instance(values_instance, provides=values_type)
    container.add_factory(
        build_service,
        provides=_VarArgsService,
        dependencies={
            values_type: values_parameter,
        },
    )

    resolved = container.resolve(_VarArgsService)

    assert isinstance(resolved, _VarArgsService)
    assert resolved.values == values_instance


def test_var_keyword_dependency_wiring_expands_mapping_dependency() -> None:
    def build_service(**options: int) -> _KwArgsService:
        return _KwArgsService(options)

    signature = inspect.signature(build_service)
    options_parameter = signature.parameters["options"]
    options_type = dict[str, int]
    options_instance = {"first": 1, "second": 2}

    container = Container()
    container.add_instance(options_instance, provides=options_type)
    container.add_factory(
        build_service,
        provides=_KwArgsService,
        dependencies={
            options_type: options_parameter,
        },
    )

    resolved = container.resolve(_KwArgsService)

    assert isinstance(resolved, _KwArgsService)
    assert resolved.options == options_instance


def test_cleanup_raises_when_only_cleanup_fails() -> None:
    class _FailingContext:
        def __enter__(self) -> _Resource:
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            raise RuntimeError("cleanup boom")

    def provide_resource() -> _FailingContext:
        return _FailingContext()

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(RuntimeError, match="cleanup boom"):
        with container.enter_scope() as request_scope:
            request_scope.resolve(_Resource)


def test_cleanup_preserves_original_exception_and_forwards_exc_info() -> None:
    received: dict[str, object | None] = {
        "exc_type": None,
        "exc_value": None,
    }

    class _TrackingContext:
        def __enter__(self) -> _Resource:
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            received["exc_type"] = exc_type
            received["exc_value"] = exc_value
            raise RuntimeError("cleanup boom")

    def provide_resource() -> _TrackingContext:
        return _TrackingContext()

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with pytest.raises(ValueError, match="original"):
        with container.enter_scope() as request_scope:
            request_scope.resolve(_Resource)
            raise ValueError("original")

    assert received["exc_type"] is ValueError
    assert isinstance(received["exc_value"], ValueError)


def test_build_root_resolver_supports_no_cleanup_mode() -> None:
    enter_calls = 0
    exit_calls = 0

    class _ManagedContext:
        def __enter__(self) -> _Resource:
            nonlocal enter_calls
            enter_calls += 1
            return _Resource()

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            traceback: TracebackType | None,
        ) -> None:
            nonlocal exit_calls
            exit_calls += 1

    def provide_resource() -> _ManagedContext:
        return _ManagedContext()

    container = Container()
    container.add_context_manager(
        provide_resource,
        provides=_Resource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    root_resolver = ResolversAssemblyCompiler().build_root_resolver(
        root_scope=Scope.APP,
        registrations=container._providers_registrations,
        cleanup_enabled=False,
    )
    request_scope = root_resolver.enter_scope()
    request_scope.resolve(_Resource)
    request_scope.__exit__(None, None, None)

    assert enter_calls == 1
    assert exit_calls == 0


def test_build_root_resolver_rebinds_globals_for_new_registrations() -> None:
    first = _Resource()
    second = _Resource()

    slot_counter = ProviderSpec.SLOT_COUNTER
    try:
        ProviderSpec.SLOT_COUNTER = 0
        first_container = Container()
        first_container.add_instance(first, provides=_Resource)

        ProviderSpec.SLOT_COUNTER = 0
        second_container = Container()
        second_container.add_instance(second, provides=_Resource)
    finally:
        ProviderSpec.SLOT_COUNTER = slot_counter

    compiler = ResolversAssemblyCompiler()
    first_resolver = compiler.build_root_resolver(
        root_scope=Scope.APP,
        registrations=first_container._providers_registrations,
    )
    assert first_resolver.resolve(_Resource) is first

    second_resolver = compiler.build_root_resolver(
        root_scope=Scope.APP,
        registrations=second_container._providers_registrations,
    )
    assert second_resolver.resolve(_Resource) is second

    # Already-cached value remains stable for the first resolver after global rebinding.
    assert first_resolver.resolve(_Resource) is first


def test_scope_chain_transitions_cover_session_request_action_and_step_paths() -> None:
    class _ActionService:
        pass

    class _StepService:
        pass

    container = Container()
    container.add(
        _SessionService,
        provides=_SessionService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _RequestService,
        provides=_RequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _ActionService,
        provides=_ActionService,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _StepService,
        provides=_StepService,
        scope=Scope.STEP,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.SESSION) as session_scope:
        assert session_scope.enter_scope(Scope.SESSION) is session_scope
        with session_scope.enter_scope() as request_scope:
            assert request_scope.enter_scope(Scope.REQUEST) is request_scope
            with request_scope.enter_scope() as action_scope:
                assert action_scope.enter_scope(Scope.ACTION) is action_scope
                with action_scope.enter_scope() as step_scope:
                    with pytest.raises(DIWireScopeMismatchError, match="Cannot enter deeper scope"):
                        step_scope.enter_scope(Scope.STEP)
                    assert isinstance(step_scope.resolve(_SessionService), _SessionService)
                    assert isinstance(step_scope.resolve(_RequestService), _RequestService)
                    assert isinstance(step_scope.resolve(_ActionService), _ActionService)
                    assert isinstance(step_scope.resolve(_StepService), _StepService)


def test_build_root_resolver_rebind_loop_preserves_each_existing_resolver_cache() -> None:
    slot_counter = ProviderSpec.SLOT_COUNTER
    try:
        ProviderSpec.SLOT_COUNTER = 0
        template_source_container = Container()
        template_source_container.add_instance(_Resource(), provides=_Resource)
        compiler = ResolversAssemblyCompiler()
        template_source_resolver = compiler.build_root_resolver(
            root_scope=Scope.APP,
            registrations=template_source_container._providers_registrations,
        )
        assert isinstance(template_source_resolver.resolve(_Resource), _Resource)

        resolvers: list[Any] = []
        instances: list[_Resource] = []
        for _ in range(4):
            ProviderSpec.SLOT_COUNTER = 0
            container = Container()
            current = _Resource()
            container.add_instance(current, provides=_Resource)
            resolver = compiler.build_root_resolver(
                root_scope=Scope.APP,
                registrations=container._providers_registrations,
            )
            assert resolver.resolve(_Resource) is current
            resolvers.append(resolver)
            instances.append(current)

        for resolver, expected in zip(resolvers, instances, strict=True):
            assert resolver.resolve(_Resource) is expected
    finally:
        ProviderSpec.SLOT_COUNTER = slot_counter



from __future__ import annotations

from types import SimpleNamespace
from typing import Any, cast

from mypy.nodes import ARG_NAMED, ARG_NAMED_OPT, ARG_OPT, ARG_POS, ARG_STAR2
from mypy.types import AnyType, TypeOfAny

from diwire._internal.integrations import mypy_plugin


class _FakeCallableType:
    def __init__(
        self,
        *,
        arg_types: list[Any],
        arg_kinds: list[Any],
        arg_names: list[str | None],
    ) -> None:
        self.arg_types = arg_types
        self.arg_kinds = arg_kinds
        self.arg_names = arg_names

    def copy_modified(
        self,
        *,
        arg_types: list[Any],
        arg_kinds: list[Any],
        arg_names: list[str | None],
    ) -> tuple[list[Any], list[Any], list[str | None]]:
        return (arg_types, arg_kinds, arg_names)


class _FakeUnionType:
    def __init__(self, items: list[Any]) -> None:
        self.items = items


class _FailingLookupApi:
    def __init__(self, error: type[BaseException]) -> None:
        self._error = error

    def named_generic_type(self, _name: str, _args: list[Any]) -> Any:
        raise self._error("lookup failed")


class _SuccessfulLookupApi:
    def __init__(self, resolved_type: Any) -> None:
        self._resolved_type = resolved_type

    def named_generic_type(self, _name: str, _args: list[Any]) -> Any:
        return self._resolved_type


def test_get_method_hook_dispatches_expected_hooks() -> None:
    plugin = cast("Any", object())

    assert (
        mypy_plugin.DIWireMypyPlugin.get_method_hook(plugin, mypy_plugin._INJECT_METHOD_FULLNAME)
        is mypy_plugin._transform_inject_method
    )
    assert (
        mypy_plugin.DIWireMypyPlugin.get_method_hook(
            plugin,
            mypy_plugin._INJECT_DECORATOR_CALL_FULLNAME,
        )
        is mypy_plugin._transform_inject_decorator_call
    )
    assert mypy_plugin.DIWireMypyPlugin.get_method_hook(plugin, "unknown.fullname") is None


def test_transform_inject_decorator_call_returns_default_when_first_argument_is_not_callable(
    monkeypatch: Any,
) -> None:
    default_return_type = object()
    ctx = SimpleNamespace(default_return_type=default_return_type)
    monkeypatch.setattr(mypy_plugin, "_first_argument_callable_type", lambda _ctx: None)

    transformed = mypy_plugin._transform_inject_decorator_call(cast("Any", ctx))

    assert transformed is default_return_type


def test_first_argument_callable_type_returns_none_for_non_callable(monkeypatch: Any) -> None:
    ctx = SimpleNamespace(arg_types=[[object()]])
    monkeypatch.setattr(mypy_plugin, "get_proper_type", lambda value: value)

    resolved = mypy_plugin._first_argument_callable_type(cast("Any", ctx))

    assert resolved is None


def test_build_precise_injected_callable_type_appends_optional_resolver(monkeypatch: Any) -> None:
    callable_type = _FakeCallableType(
        arg_types=["injected_param"],
        arg_kinds=[ARG_POS],
        arg_names=["service"],
    )
    monkeypatch.setattr(
        mypy_plugin, "_unwrap_injected_parameter_type", lambda _value: ("Service", True)
    )
    monkeypatch.setattr(mypy_plugin, "_resolver_protocol_type", lambda _ctx: "ResolverProtocol")

    transformed = mypy_plugin._build_precise_injected_callable_type(
        ctx=cast("Any", SimpleNamespace()),
        callable_type=cast("Any", callable_type),
    )

    assert transformed == (
        ["Service", "ResolverProtocol"],
        [ARG_OPT, ARG_NAMED_OPT],
        ["service", "diwire_resolver"],
    )


def test_build_precise_injected_callable_type_keeps_existing_resolver_kwarg(
    monkeypatch: Any,
) -> None:
    callable_type = _FakeCallableType(
        arg_types=["ResolverProtocol"],
        arg_kinds=[ARG_NAMED_OPT],
        arg_names=["diwire_resolver"],
    )
    monkeypatch.setattr(
        mypy_plugin, "_unwrap_injected_parameter_type", lambda value: (value, False)
    )

    transformed = mypy_plugin._build_precise_injected_callable_type(
        ctx=cast("Any", SimpleNamespace()),
        callable_type=cast("Any", callable_type),
    )

    assert transformed == (
        ["ResolverProtocol"],
        [ARG_NAMED_OPT],
        ["diwire_resolver"],
    )


def test_build_precise_injected_callable_type_inserts_resolver_before_kwargs(
    monkeypatch: Any,
) -> None:
    callable_type = _FakeCallableType(
        arg_types=["required", "kwargs"],
        arg_kinds=[ARG_POS, ARG_STAR2],
        arg_names=["required", None],
    )
    monkeypatch.setattr(
        mypy_plugin, "_unwrap_injected_parameter_type", lambda value: (value, False)
    )
    monkeypatch.setattr(mypy_plugin, "_resolver_protocol_type", lambda _ctx: "ResolverProtocol")

    transformed = mypy_plugin._build_precise_injected_callable_type(
        ctx=cast("Any", SimpleNamespace()),
        callable_type=cast("Any", callable_type),
    )

    assert transformed == (
        ["required", "ResolverProtocol", "kwargs"],
        [ARG_POS, ARG_NAMED_OPT, ARG_STAR2],
        ["required", "diwire_resolver", None],
    )


def test_unwrap_injected_parameter_type_rejects_non_duplicate_union_shapes(
    monkeypatch: Any,
) -> None:
    original_type = object()
    monkeypatch.setattr(mypy_plugin, "UnionType", _FakeUnionType)
    monkeypatch.setattr(mypy_plugin, "get_proper_type", lambda _value: _FakeUnionType([1, 2, 3]))

    unwrapped, is_injected = mypy_plugin._unwrap_injected_parameter_type(cast("Any", original_type))

    assert unwrapped is original_type
    assert is_injected is False


def test_unwrap_injected_parameter_type_rejects_non_matching_duplicate_union(
    monkeypatch: Any,
) -> None:
    original_type = object()
    monkeypatch.setattr(mypy_plugin, "UnionType", _FakeUnionType)
    monkeypatch.setattr(mypy_plugin, "get_proper_type", lambda _value: _FakeUnionType([1, 2]))
    monkeypatch.setattr(mypy_plugin, "is_same_type", lambda _left, _right: False)

    unwrapped, is_injected = mypy_plugin._unwrap_injected_parameter_type(cast("Any", original_type))

    assert unwrapped is original_type
    assert is_injected is False


def test_to_optional_argument_kind_maps_named_and_passthrough() -> None:
    assert mypy_plugin._to_optional_argument_kind(ARG_NAMED) == ARG_NAMED_OPT
    assert mypy_plugin._to_optional_argument_kind(ARG_NAMED_OPT) == ARG_NAMED_OPT


def test_resolver_protocol_type_returns_any_when_lookup_fails() -> None:
    for error in (AssertionError, KeyError):
        ctx = SimpleNamespace(api=_FailingLookupApi(error))
        resolved = mypy_plugin._resolver_protocol_type(cast("Any", ctx))
        assert isinstance(resolved, AnyType)
        assert resolved.type_of_any == TypeOfAny.from_error


def test_resolver_protocol_type_returns_named_type_when_lookup_succeeds() -> None:
    expected_type = object()
    ctx = SimpleNamespace(api=_SuccessfulLookupApi(expected_type))

    resolved = mypy_plugin._resolver_protocol_type(cast("Any", ctx))

    assert resolved is expected_type



from dataclasses import dataclass
from types import TracebackType

from starlette.datastructures import State
from starlette.requests import Request
from starlette.websockets import WebSocket
from typing_extensions import Self


class CustomState(State):
    pass


@dataclass
class RequestBasedService:
    request: Request

    def work(self) -> str:
        return f"Request path: {self.request.url.path}"


@dataclass
class RequestWithCustomStateService:
    request: Request[CustomState]

    def work(self) -> str:
        return f"Custom state value: {self.request.url.path}"


@dataclass
class WebSocketBasedService:
    websocket: WebSocket

    def work(self) -> str:
        return f"WebSocket path: {self.websocket.url.path}"


@dataclass
class CMService:
    def work(self) -> str:
        return "CMService working"

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        print("CMService __exit__")



from __future__ import annotations

from typing import Any

import rodi
from dishka import Provider
from wireup import injectable

from diwire import Lifetime
from tests.benchmarks.dishka_helpers import DishkaBenchmarkScope, make_dishka_benchmark_container
from tests.benchmarks.helpers import make_diwire_benchmark_container, run_benchmark
from tests.benchmarks.wireup_helpers import make_wireup_benchmark_container


@injectable(lifetime="transient")
class _TransientService:
    pass


def test_benchmark_diwire_resolve_transient(benchmark: Any) -> None:
    container = make_diwire_benchmark_container()
    container.add(_TransientService, lifetime=Lifetime.TRANSIENT)
    container.compile()
    first = container.resolve(_TransientService)
    second = container.resolve(_TransientService)
    assert first is not second

    def bench_diwire_transient() -> None:
        _ = container.resolve(_TransientService)

    run_benchmark(benchmark, bench_diwire_transient)


def test_benchmark_rodi_resolve_transient(benchmark: Any) -> None:
    rodi_container = rodi.Container()
    rodi_container.add_transient(_TransientService)
    services = rodi_container.build_provider()
    first = services.get(_TransientService)
    second = services.get(_TransientService)
    assert first is not second

    def bench_rodi_transient() -> None:
        _ = services.get(_TransientService)

    run_benchmark(benchmark, bench_rodi_transient)


def test_benchmark_dishka_resolve_transient(benchmark: Any) -> None:
    provider = Provider(scope=DishkaBenchmarkScope.APP)
    provider.provide(_TransientService, scope=DishkaBenchmarkScope.APP, cache=False)
    container = make_dishka_benchmark_container(provider)
    first = container.get(_TransientService)
    second = container.get(_TransientService)
    assert first is not second

    def bench_dishka_transient() -> None:
        _ = container.get(_TransientService)

    run_benchmark(benchmark, bench_dishka_transient)


def test_benchmark_wireup_resolve_transient(benchmark: Any) -> None:
    container = make_wireup_benchmark_container(_TransientService)
    with container.enter_scope() as scope:
        first = scope.get(_TransientService)
        second = scope.get(_TransientService)
        assert first is not second

        def bench_wireup_transient() -> None:
            _ = scope.get(_TransientService)

        run_benchmark(benchmark, bench_wireup_transient)



"""Focused example: ``DIWireDependencyNotRegisteredError``."""

from __future__ import annotations

from diwire import Container, DependencyRegistrationPolicy, MissingPolicy
from diwire.exceptions import DIWireDependencyNotRegisteredError


class MissingDependency:
    pass


def main() -> None:
    container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )

    try:
        container.resolve(MissingDependency)
    except DIWireDependencyNotRegisteredError as error:
        error_name = type(error).__name__

    print(f"missing={error_name}")  # => missing=DIWireDependencyNotRegisteredError


if __name__ == "__main__":
    main()



"""Focused example: resolve-time autoregistration of a dependency chain."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container


class Leaf:
    pass


@dataclass(slots=True)
class Branch:
    leaf: Leaf


@dataclass(slots=True)
class Root:
    branch: Branch


def main() -> None:
    container = Container()
    resolved = container.resolve(Root)
    print(
        f"autoregister_chain={isinstance(resolved.branch.leaf, Leaf)}",
    )  # => autoregister_chain=True


if __name__ == "__main__":
    main()



from __future__ import annotations

from celery import Celery, Task

from diwire import Container, Injected, Lifetime, Scope, resolver_context
from tests.e2e.celery.setup.config import CeleryE2ESettings
from tests.e2e.celery.setup.services import AdderService, ScopedToken, build_scoped_token

container = Container()
container.add(AdderService, provides=AdderService)
container.add_factory(
    build_scoped_token,
    provides=ScopedToken,
    scope=Scope.REQUEST,
    lifetime=Lifetime.SCOPED,
)

settings = container.resolve(CeleryE2ESettings)
celery_app = Celery(
    "diwire-celery-e2e",
    broker=settings.broker_url,
    backend=settings.result_backend,
)
celery_app.conf.update(
    task_default_queue=settings.queue,
    task_serializer="json",
    result_serializer="json",
    accept_content=["json"],
)


@celery_app.task(name="tests.e2e.celery.tasks.add")
@resolver_context.inject(scope=Scope.REQUEST)
def add_task(left: int, right: int, adder: Injected[AdderService]) -> int:
    return adder.add(left, right)


@celery_app.task(name="tests.e2e.celery.tasks.scope")
@resolver_context.inject(scope=Scope.REQUEST)
def scope_task(
    first: Injected[ScopedToken],
    second: Injected[ScopedToken],
) -> dict[str, str]:
    return {"first": first.value, "second": second.value}


@celery_app.task(bind=True, name="tests.e2e.celery.tasks.bound")
@resolver_context.inject(scope=Scope.REQUEST)
def bound_task(task: Task, value: int, adder: Injected[AdderService]) -> str:
    return f"{task.name}:{adder.add(value, 1)}"


app = celery_app



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING

import pytest
from typing_extensions import Self

from diwire import Container, Injected, Lifetime, Scope, resolver_context

if TYPE_CHECKING:
    from flask import Flask, Request
    from flask.testing import FlaskClient

_cleanup_paths: list[str] = []


def _latest_cleanup_path() -> str | None:
    if not _cleanup_paths:
        return None
    return _cleanup_paths[-1]


@dataclass
class _RequestPathService:
    request: Request

    def path(self) -> str:
        return self.request.path


@dataclass
class _CMService:
    request: Request

    def work(self) -> str:
        return "CMService working"

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        _cleanup_paths.append(self.request.path)


@pytest.fixture()
def app() -> Flask:
    pytest.importorskip("flask")

    from flask import Flask, Request

    from diwire.integrations.flask import add_request_context

    globals()["Request"] = Request
    _cleanup_paths.clear()

    container = Container()
    add_request_context(container)
    container.add(
        _RequestPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_context_manager(_CMService, scope=Scope.REQUEST)

    app = Flask(__name__)

    @app.get("/request/direct")
    @resolver_context.inject(scope=Scope.REQUEST)
    def request_direct(request: Injected[Request]) -> dict[str, str]:
        return {"path": request.path}

    @app.get("/request/service")
    @resolver_context.inject(scope=Scope.REQUEST)
    def request_service(service: Injected[_RequestPathService]) -> dict[str, str]:
        return {"path": service.path()}

    @app.get("/services/cm-service")
    @resolver_context.inject(scope=Scope.REQUEST)
    def cm_service(service: Injected[_CMService]) -> dict[str, str]:
        return {"message": service.work()}

    @app.get("/services/cm-service/cleanup")
    def cm_service_cleanup() -> dict[str, str | None]:
        return {"path": _latest_cleanup_path()}

    return app


@pytest.fixture()
def client(app: Flask) -> Iterator[FlaskClient]:
    with app.test_client() as test_client:
        yield test_client


def test_request_resolve_for_http_endpoint(client: FlaskClient) -> None:
    response = client.get("/request/direct")
    assert response.status_code == 200
    assert response.json == {"path": "/request/direct"}


def test_request_resolve_in_service_for_http_endpoint(client: FlaskClient) -> None:
    response = client.get("/request/service")
    assert response.status_code == 200
    assert response.json == {"path": "/request/service"}


def test_context_manager_service_validates_request_scope_cleanup_path(client: FlaskClient) -> None:
    response = client.get("/services/cm-service")
    assert response.status_code == 200
    assert response.json == {"message": "CMService working"}

    cleanup_response = client.get("/services/cm-service/cleanup")
    assert cleanup_response.status_code == 200
    assert cleanup_response.json == {"path": "/services/cm-service"}



from __future__ import annotations

import inspect
from collections.abc import Generator
from contextlib import contextmanager

import pytest

from diwire import Container
from diwire.exceptions import DIWireInvalidProviderSpecError, DIWireProviderDependencyInferenceError


class UntypedDependency:
    pass


class TypedDependency:
    pass


class Service:
    pass


class ConcreteService(Service):
    def __init__(self, dep: TypedDependency) -> None:
        self.dep = dep


def test_explicit_dependencies_bypass_inference() -> None:
    def build_service(  # type: ignore[no-untyped-def]
        raw_dependency,
        typed_dependency: TypedDependency,
    ) -> Service:
        return Service()

    signature = inspect.signature(build_service)
    dependencies = {
        UntypedDependency: signature.parameters["raw_dependency"],
        TypedDependency: signature.parameters["typed_dependency"],
    }

    container = Container()
    container.add_factory(build_service, provides=Service, dependencies=dependencies)

    provider_spec = container._providers_registrations.get_by_type(Service)
    assert [dependency.parameter.name for dependency in provider_spec.dependencies] == [
        "raw_dependency",
        "typed_dependency",
    ]
    assert [dependency.provides for dependency in provider_spec.dependencies] == [
        UntypedDependency,
        TypedDependency,
    ]


def test_missing_explicit_dependencies_raises_inference_error() -> None:
    def build_service(raw_dependency) -> Service:  # type: ignore[no-untyped-def]
        return Service()

    container = Container()

    with pytest.raises(DIWireProviderDependencyInferenceError, match="raw_dependency"):
        container.add_factory(build_service, provides=Service)


def test_direct_registration_accepts_explicit_dependencies() -> None:
    def build_service(raw_dependency) -> Service:  # type: ignore[no-untyped-def]
        return Service()

    signature = inspect.signature(build_service)
    dependencies = {
        UntypedDependency: signature.parameters["raw_dependency"],
    }

    container = Container()
    container.add_factory(build_service, provides=Service, dependencies=dependencies)

    provider_spec = container._providers_registrations.get_by_type(Service)
    assert [dependency.parameter.name for dependency in provider_spec.dependencies] == [
        "raw_dependency",
    ]
    assert [dependency.provides for dependency in provider_spec.dependencies] == [UntypedDependency]


def test_explicit_dependencies_reject_unknown_parameter() -> None:
    def build_service(dep: TypedDependency) -> Service:
        return Service()

    dependencies = {
        TypedDependency: inspect.Parameter(
            "unknown_param",
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ),
    }

    container = Container()

    with pytest.raises(DIWireInvalidProviderSpecError, match="unknown parameter"):
        container.add_factory(build_service, provides=Service, dependencies=dependencies)


def test_explicit_dependencies_reject_duplicates() -> None:
    def build_service(dep: TypedDependency) -> Service:
        return Service()

    parameter = inspect.signature(build_service).parameters["dep"]
    dependencies = {
        TypedDependency: parameter,
        UntypedDependency: parameter,
    }

    container = Container()

    with pytest.raises(DIWireInvalidProviderSpecError, match="duplicated"):
        container.add_factory(build_service, provides=Service, dependencies=dependencies)


def test_explicit_dependencies_reject_kind_mismatch() -> None:
    def build_service(dep: TypedDependency) -> Service:
        return Service()

    dependencies = {
        TypedDependency: inspect.Parameter("dep", inspect.Parameter.KEYWORD_ONLY),
    }

    container = Container()

    with pytest.raises(DIWireInvalidProviderSpecError, match="has kind"):
        container.add_factory(build_service, provides=Service, dependencies=dependencies)


def test_explicit_dependencies_for_concrete_type_are_validated() -> None:
    signature = inspect.signature(ConcreteService.__init__)
    dependencies = {
        TypedDependency: signature.parameters["dep"],
    }

    container = Container()
    container.add(ConcreteService, provides=Service, dependencies=dependencies)

    provider_spec = container._providers_registrations.get_by_type(Service)
    assert [dependency.parameter.name for dependency in provider_spec.dependencies] == ["dep"]
    assert [dependency.provides for dependency in provider_spec.dependencies] == [TypedDependency]


def test_explicit_dependencies_for_generator_are_validated() -> None:
    def build_service(dep: TypedDependency) -> Generator[Service, None, None]:
        try:
            yield Service()
        finally:
            pass

    signature = inspect.signature(build_service)
    dependencies = {
        TypedDependency: signature.parameters["dep"],
    }

    container = Container()
    container.add_generator(build_service, provides=Service, dependencies=dependencies)

    provider_spec = container._providers_registrations.get_by_type(Service)
    assert [dependency.parameter.name for dependency in provider_spec.dependencies] == ["dep"]
    assert [dependency.provides for dependency in provider_spec.dependencies] == [TypedDependency]


def test_explicit_dependencies_for_context_manager_are_validated() -> None:
    @contextmanager
    def build_service(dep: TypedDependency) -> Generator[Service, None, None]:
        yield Service()

    signature = inspect.signature(build_service)
    dependencies = {
        TypedDependency: signature.parameters["dep"],
    }

    container = Container()
    container.add_context_manager(
        build_service,
        provides=Service,
        dependencies=dependencies,
    )

    provider_spec = container._providers_registrations.get_by_type(Service)
    assert [dependency.parameter.name for dependency in provider_spec.dependencies] == ["dep"]
    assert [dependency.provides for dependency in provider_spec.dependencies] == [TypedDependency]



from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

import pytest

from diwire import Container, Injected, Lifetime, Scope, resolver_context

if TYPE_CHECKING:
    from fastapi import FastAPI
    from fastapi.testclient import TestClient
    from starlette.datastructures import State
    from starlette.requests import Request
    from starlette.websockets import WebSocket


class _RequestPathService:
    def __init__(self, request: Request[State]) -> None:
        self._request = request

    def path(self) -> str:
        return self._request.url.path


class _WebSocketPathService:
    def __init__(self, websocket: WebSocket) -> None:
        self._websocket = websocket

    def path(self) -> str:
        return self._websocket.url.path


@pytest.fixture()
def app() -> FastAPI:
    pytest.importorskip("fastapi")

    from fastapi import FastAPI
    from starlette.datastructures import State
    from starlette.requests import Request
    from starlette.websockets import WebSocket

    from diwire.integrations.fastapi import RequestContextMiddleware, add_request_context

    # diwire resolves dependencies via runtime type hints; populate module globals so
    # forward references like `Request[State]` and `WebSocket` can be evaluated.
    globals()["Request"] = Request
    globals()["State"] = State
    globals()["WebSocket"] = WebSocket

    container = Container()
    add_request_context(container)
    container.add(
        _RequestPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _WebSocketPathService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    app = FastAPI()
    app.add_middleware(RequestContextMiddleware)

    @app.get("/request/direct")
    @resolver_context.inject(scope=Scope.REQUEST)
    def request_direct(request: Injected[Request[State]]) -> dict[str, str]:
        return {"path": request.url.path}

    @app.get("/request/service")
    @resolver_context.inject(scope=Scope.REQUEST)
    def request_service(service: Injected[_RequestPathService]) -> dict[str, str]:
        return {"path": service.path()}

    @app.websocket("/websocket/direct")
    @resolver_context.inject(scope=Scope.REQUEST)
    async def websocket_direct(
        websocket: WebSocket,
        resolved_websocket: Injected[WebSocket],
    ) -> None:
        await websocket.accept()
        await websocket.send_json({"path": resolved_websocket.url.path})
        await websocket.close()

    @app.websocket("/websocket/service")
    @resolver_context.inject(scope=Scope.REQUEST)
    async def websocket_service(
        websocket: WebSocket,
        service: Injected[_WebSocketPathService],
    ) -> None:
        await websocket.accept()
        await websocket.send_json({"path": service.path()})
        await websocket.close()

    return app


@pytest.fixture()
def client(app: FastAPI) -> Iterator[TestClient]:
    pytest.importorskip("fastapi")

    from fastapi.testclient import TestClient

    with TestClient(app) as test_client:
        yield test_client


def test_request_resolve_for_http_endpoint(client: TestClient) -> None:
    response = client.get("/request/direct")
    assert response.status_code == 200
    assert response.json() == {"path": "/request/direct"}


def test_request_resolve_in_service_for_http_endpoint(client: TestClient) -> None:
    response = client.get("/request/service")
    assert response.status_code == 200
    assert response.json() == {"path": "/request/service"}


def test_websocket_resolve_for_websocket_endpoint(client: TestClient) -> None:
    with client.websocket_connect("/websocket/direct") as websocket:
        payload = websocket.receive_json()
    assert payload == {"path": "/websocket/direct"}


def test_websocket_resolve_in_service_for_websocket_endpoint(client: TestClient) -> None:
    with client.websocket_connect("/websocket/service") as websocket:
        payload = websocket.receive_json()
    assert payload == {"path": "/websocket/service"}



from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from mypy.plugin import Plugin


__all__ = ["plugin"]


def plugin(version: str) -> type[Plugin]:
    """Return the DIWire mypy plugin class for mypy's plugin loader.

    Args:
        version: Mypy version string provided by mypy's plugin loader.

    Returns:
        DIWire mypy plugin class.

    """
    _ = version
    module = importlib.import_module("diwire._internal.integrations.mypy_plugin")
    return cast("type[Plugin]", module.DIWireMypyPlugin)



"""Class-based context manager registration with inferred managed type.

This module demonstrates:

1. Registering a class context manager directly via ``add_context_manager(Service)``.
2. Inferring ``provides`` from ``Service.__enter__``.
3. Request-scoped caching behavior (same instance within one request scope).
"""

from __future__ import annotations

from types import TracebackType

from typing_extensions import Self

from diwire import Container, Lifetime, Scope


class Service:
    def __init__(self) -> None:
        print("new Service")  # => new Service

    def __enter__(self) -> Self:
        print("Entering Service context")  # => Entering Service context
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        print("Exiting Service context")  # => Exiting Service context


def main() -> None:
    container = Container()
    container.add_context_manager(
        Service,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST) as request_scope:
        service_1 = request_scope.resolve(Service)
        service_2 = request_scope.resolve(Service)
        same_instance = service_1 is service_2
    print(f"same_instance={same_instance}")  # => same_instance=True


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
from uuid import uuid4


@dataclass(frozen=True)
class AdderService:
    def add(self, left: int, right: int) -> int:
        return left + right


@dataclass(frozen=True)
class ScopedToken:
    value: str


def build_scoped_token() -> ScopedToken:
    return ScopedToken(value=uuid4().hex)



from __future__ import annotations

import pytest

pytest.importorskip("aiohttp")

from aiohttp import web
from aiohttp.test_utils import make_mocked_request

import diwire._internal.integrations.aiohttp as aiohttp_integration
from diwire import Container
from diwire.exceptions import DIWireIntegrationError


def test_get_request_raises_when_context_is_missing() -> None:
    with pytest.raises(DIWireIntegrationError, match="Request context not available"):
        aiohttp_integration.get_request()


@pytest.mark.asyncio
async def test_request_context_middleware_sets_context_for_handler() -> None:
    request = make_mocked_request("GET", "/middleware/context")

    async def handler(current_request: web.Request) -> web.StreamResponse:
        resolved_request = aiohttp_integration.get_request()
        assert resolved_request is current_request
        return web.Response(text="ok")

    response = await aiohttp_integration.request_context_middleware(request, handler)

    assert response.status == 200


@pytest.mark.asyncio
async def test_request_context_middleware_resets_context_after_success() -> None:
    request = make_mocked_request("GET", "/middleware/success")

    async def handler(_request: web.Request) -> web.StreamResponse:
        assert aiohttp_integration.get_request().path == "/middleware/success"
        return web.Response(text="ok")

    await aiohttp_integration.request_context_middleware(request, handler)

    with pytest.raises(DIWireIntegrationError, match="Request context not available"):
        aiohttp_integration.get_request()


@pytest.mark.asyncio
async def test_request_context_middleware_resets_context_after_exception() -> None:
    request = make_mocked_request("GET", "/middleware/error")

    async def handler(_request: web.Request) -> web.StreamResponse:
        assert aiohttp_integration.get_request().path == "/middleware/error"
        msg = "handler failed"
        raise RuntimeError(msg)

    with pytest.raises(RuntimeError, match="handler failed"):
        await aiohttp_integration.request_context_middleware(request, handler)

    with pytest.raises(DIWireIntegrationError, match="Request context not available"):
        aiohttp_integration.get_request()


def test_add_request_context_registers_factory_for_container_resolution() -> None:
    container = Container()
    aiohttp_integration.add_request_context(container)

    request = make_mocked_request("GET", "/container/request")
    token = aiohttp_integration._request_context.set(request)
    try:
        resolved_request = container.resolve(web.Request)
        assert resolved_request is request
    finally:
        aiohttp_integration._request_context.reset(token)
